import numpy as np 
import scipy.special 
from . import ParameterError
from .torch_numpy_ops import get_npt

EPS64 = float(np.finfo(np.float64).eps)
    
def insert_batch_dims(param, ndims, k):
    ones = [1]*ndims
    return param.reshape(list(param.shape[:k])+ones+list(param.shape[k:]))

def tf_exp(x):
    npt = get_npt(x)
    return npt.exp(x) 

def tf_exp_inv(x):
    npt = get_npt(x)
    return npt.log(x) 

def tf_exp_eps(x):
    return tf_exp(x)+EPS64 

def tf_exp_eps_inv(x):
    return tf_exp_inv(x-EPS64) 

def tf_square(x):
    return x**2 

def tf_square_inv(x):
    npt = get_npt(x)
    return npt.sqrt(x) 

def tf_square_eps(x):
    return tf_square(x)+EPS64 

def tf_square_eps_inv(x):
    return tf_square_inv(x-EPS64)

def tf_explinear(x):
    npt = get_npt(x)
    if npt==np:
        return -scipy.special.log_expit(-x)
    else:
        return -npt.nn.functional.logsigmoid(-x)

def tf_explinear_inv(x):
    npt = get_npt(x)
    return npt.where(x<34,npt.log(npt.expm1(x)),x)

def tf_explinear_eps(x):
    return tf_explinear(x)+EPS64

def tf_explinear_eps_inv(x):
    return tf_explinear_inv(x-EPS64)

def tf_identity(x):
    return x 

def parse_assign_param(pname, param, shape_param, requires_grad_param, tfs_param, endsize_ops, constraints, torchify, npt, nptkwargs):
    if np.isscalar(param):
        param = param*npt.ones(shape_param,**nptkwargs)
    else:
        if torchify:
            if not isinstance(param,npt.Tensor):
                param = npt.tensor(param)
            param = npt.atleast_1d(param)
            assert isinstance(param,npt.Tensor), "%s must be a scalar or torch.Tensor"%pname
        else: 
            if not isinstance(param,npt.ndarray):
                param = npt.array(param)
            param = npt.atleast_1d(param)
            assert isinstance(param,npt.ndarray), "%s must be a scalar or np.ndarray"%pname
    shape_param = list(param.shape)
    assert len(shape_param)>=1, "invalid shape_%s = %s"%(pname,str(shape_param))
    assert len(tfs_param)==2, "tfs_scale should be a tuple of length 2"
    assert callable(tfs_param[0]), "tfs_scale[0] should be a callable e.g. torch.log"
    assert callable(tfs_param[1]), "tfs_scale[1] should be a callable e.g. torch.exp"
    raw_param = tfs_param[0](param)
    if torchify:
        assert isinstance(requires_grad_param,bool)
        if requires_grad_param:
            raw_param = 1.*raw_param
        raw_param = npt.nn.Parameter(raw_param,requires_grad=requires_grad_param)
    assert shape_param[-1] in endsize_ops, "%s not in %s"%(str(shape_param[-1]),str(endsize_ops))
    if "POSITIVE" in constraints: 
        assert (param>0).all(), "%s must be positive"%pname
    if "NON-NEGATIVE" in constraints: 
        assert (param>=0).all(), "%s must be non-negative"%pname
    if "INTEGER" in constraints:
        assert (param%1==0).all(), "%s must be integers"%pname
    return raw_param



from .ft import fftbr,ifftbr,fwht,omega_fftbr,omega_fwht
try:
    import torch 
    from .ft_pytorch import (
        fftbr_torch,
        ifftbr_torch,
        fwht_torch,
        omega_fftbr_torch,
        omega_fwht_torch,
    )
except ImportError:
    def fftbr_torch(*args, **kwargs):
        raise Exception("fftbr_torch requires torch but no installation found")
    def ifftbr_torch(*args, **kwargs):
        raise Exception("ifftbr_torch requires torch but no installation found") 
    def fwht_torch(*args, **kwargs):
        raise Exception("fwht_torch requires torch but no installation found") 
    def omega_fftbr_torch(*args, **kwargs):
        raise Exception("omega_fftbr_torch requires torch but no installation found") 
    def omega_fwht_torch(*args, **kwargs):
        raise Exception("omega_fwht_torch requires torch but no installation found") 


from .abstract_discrete_distribution import AbstractDiscreteDistribution
from .iid_std_uniform import IIDStdUniform
from .lattice import Lattice
from .digital_net_b2 import DigitalNetB2
from .halton import Halton

DiscreteDistribution = AbstractDiscreteDistribution
_DiscreteDistribution = AbstractDiscreteDistribution
Sobol = DigitalNetB2
DigitalNet = DigitalNetB2
Net = DigitalNetB2




from .abstract_integrand import AbstractIntegrand
from .financial_option import (
    FinancialOption,
    AsianOption,
    EuropeanOption,
    BarrierOption,
    LookbackOption,
    DigitalOption)
from .keister import Keister
from .linear0 import Linear0
from .custom_fun import CustomFun
from .box_integral import BoxIntegral
from .sensitivity_indices import SensitivityIndices
from .ishigami import Ishigami
from .bayesian_lr_coeffs import BayesianLRCoeffs
from .umbridge_wrapper import UMBridgeWrapper
from .genz import Genz
from .sin1d import Sin1d
from .hartmann6d import Hartmann6d
from .fourbranch2d import FourBranch2d
from .multimodal2d import Multimodal2d

Integrand = AbstractIntegrand
_Integrand = AbstractIntegrand
SobolIndices = SensitivityIndices
CustomIntegrand = CustomFun
UserFun = CustomFun
UserIntegrand = CustomFun



import qmctoolscl 
import numpy as np 
import scipy.fft 
import itertools


def fftbr(x):
    r"""
    1 dimensional Bit-Reversed-Order (BRO) Fast Fourier Transform (FFT) along the last dimension. 
    Requires the last dimension of x is already in BRO, so we can skip the first step of the decimation-in-time FFT. 
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = rng.random(8)+1j*rng.random(8)
        >>> with np.printoptions(precision=2):
        ...     fftbr(x)
        array([ 1.07+1.14j, -0.32+0.26j, -0.27+0.22j, -0.27-0.34j,  0.1 +0.16j,
               -0.15+0.17j,  0.5 +0.24j,  0.06-0.02j])
        >>> fftbr(rng.random((2,3,4,5,8))+1j*rng.random((2,3,4,5,8))).shape
        (2, 3, 4, 5, 8)

    Args:
        x (np.ndarray): Array of samples at which to run BRO-FFT.
    
    Returns:
        y (np.ndarray): BRO-FFT values.
    """
    n = x.shape[-1]
    assert n&(n-1)==0 # require n is a power of 2
    m = int(np.log2(n))
    shape = list(x.shape)
    ndim = x.ndim
    twos = [2]*m
    xs = x.reshape(shape[:-1]+twos)
    pdims = tuple(itertools.chain(range(ndim-1),range(m+ndim-2,ndim-2,-1)))#[i for i in range(ndim-1)]+[i+ndim-1 for i in range(m-1,-1,-1)]
    xrf = np.moveaxis(xs,np.arange(len(pdims)),pdims)
    xr = np.ascontiguousarray(xrf).reshape(shape)
    return scipy.fft.fft(xr,norm="ortho")

def ifftbr(x):
    r"""
    1 dimensional Bit-Reversed-Order (BRO) Inverse Fast Fourier Transform (IFFT) along the last dimension.  
    Outputs an array in bit-reversed order, so we can skip the last step of the decimation-in-time IFFT.  
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = rng.random(8)+1j*rng.random(8)
        >>> with np.printoptions(precision=2):
        ...     ifftbr(x)
        array([ 1.07+1.14j, -0.29-0.22j,  0.3 +0.06j, -0.09+0.02j,  0.03+0.54j,
               -0.04-0.33j, -0.19+0.26j, -0.08+0.36j])
        >>> ifftbr(rng.random((2,3,4,5,8))+1j*rng.random((2,3,4,5,8))).shape
        (2, 3, 4, 5, 8)
    
    Args:
        x (np.ndarray): Array of samples at which to run BRO-IFFT.
    
    Returns:
        y (np.ndarray): BRO-IFFT values.
    """
    n = x.shape[-1]
    assert n&(n-1)==0 # require n is a power of 2
    m = int(np.log2(n))
    shape = list(x.shape)
    ndim = x.ndim
    twos = [2]*m
    x = scipy.fft.ifft(x,norm="ortho")
    xs = x.reshape(shape[:-1]+twos)
    pdims = tuple(itertools.chain(range(ndim-1),range(m+ndim-2,ndim-2,-1)))
    xrf = np.moveaxis(xs,np.arange(len(pdims)),pdims)
    xr = np.ascontiguousarray(xrf).reshape(shape)
    return xr

def fwht(x):
    r"""
    1 dimensional Fast Walsh Hadamard Transform (FWHT) along the last dimension.  
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> with np.printoptions(precision=2):
        ...     fwht(rng.random(8))
        array([ 1.07, -0.29,  0.12,  0.08,  0.1 , -0.45,  0.1 , -0.03])
        >>> fwht(rng.random((2,3,4,5,8))).shape
        (2, 3, 4, 5, 8)

    Args:
        x (np.ndarray): Array of samples at which to run FWHT.
    
    Returns:
        y (np.ndarray): FWHT values.
    """
    y = x.copy()+0.
    n = x.shape[-1]
    if n<=1: return y
    assert n&(n-1)==0 # require n is a power of 2
    m = int(np.log2(n))
    it = np.arange(n,dtype=np.int64).reshape([2]*m) # 2 x 2 x ... x 2 array (size 2^m)
    idx0 = [slice(None)]*(m-1)+[0]
    idx1 = [slice(None)]*(m-1)+[1]
    for k in range(m):
        eps0 = it[tuple(idx0[-(k+1):])].flatten()
        eps1 = it[tuple(idx1[-(k+1):])].flatten()
        y0,y1 = y[...,eps0],y[...,eps1]
        y[...,eps0],y[...,eps1] = (y0+y1)/np.sqrt(2),(y0-y1)/np.sqrt(2)
    return y

def omega_fwht(m):
    r"""
    A useful when efficiently updating FWHT values after doubling the sample size.  

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> m = 3
        >>> x1 = rng.random((3,5,7,2**m))
        >>> x2 = rng.random((3,5,7,2**m))
        >>> x = np.concatenate([x1,x2],axis=-1)
        >>> ytrue = fwht(x)
        >>> omega = omega_fwht(m)
        >>> y1 = fwht(x1) 
        >>> y2 = fwht(x2) 
        >>> y = np.concatenate([y1+omega*y2,y1-omega*y2],axis=-1)/np.sqrt(2)
        >>> np.allclose(y,ytrue)
        True
    
    Args:
        m (int): Size $2^m$ output. 
    
    Returns:
        y (np.ndarray): $\left(1\right)_{k=0}^{2^m}$. 
    """
    return np.ones(2**m)

def omega_fftbr(m):
    r"""
    A useful when efficiently updating FFT values after doubling the sample size. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> m = 3
        >>> x1 = rng.random((3,5,7,2**m))+1j*rng.random((3,5,7,2**m))
        >>> x2 = rng.random((3,5,7,2**m))+1j*rng.random((3,5,7,2**m))
        >>> x = np.concatenate([x1,x2],axis=-1)
        >>> ytrue = fftbr(x)
        >>> omega = omega_fftbr(m)
        >>> y1 = fftbr(x1) 
        >>> y2 = fftbr(x2) 
        >>> y = np.concatenate([y1+omega*y2,y1-omega*y2],axis=-1)/np.sqrt(2)
        >>> np.allclose(y,ytrue)
        True
    
    Args:
        m (int): Size $2^m$ output. 
    
    Returns:
        y (np.ndarray): $\left(e^{- \pi \mathrm{i} k / 2^m}\right)_{k=0}^{2^m}$. 
    """
    return np.exp(-np.pi*1j*np.arange(2**m)/2**m)



from .discrete_distribution import *
from .true_measure import *
from .integrand import *
from .stopping_criterion import *
from .kernel import (
    KernelGaussian,
    KernelSquaredExponential,
    KernelRationalQuadratic,
    KernelMatern12,
    KernelMatern32,
    KernelMatern52,
    KernelShiftInvar,
    KernelSI,
    KernelShiftInvarCombined,
    KernelSICombined,
    KernelDigShiftInvar,
    KernelDSI,
    KernelDigShiftInvarAdaptiveAlpha,
    KernelDSIAA,
    KernelDigShiftInvarCombined,
    KernelDSICombined,
    KernelMultiTask,
    KernelMultiTaskDerivs,

)
from .fast_transform import (
    fftbr,
    ifftbr,
    fwht,
    omega_fftbr,
    omega_fwht,
    fftbr_torch,
    ifftbr_torch,
    fwht_torch,
    omega_fftbr_torch,
    omega_fwht_torch,
)
from .util import plot_proj,mlmc_test

name = "qmcpy"
__version__ = "2.0.0.1c3"



from .abstract_true_measure import AbstractTrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
from ..discrete_distribution import DigitalNetB2
import numpy as np
import scipy.stats
from typing import Union


class SciPyWrapper(AbstractTrueMeasure):
    r"""
    Multivariate distribution with independent marginals from [`scipy.stats`](https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions).

    Examples:
        >>> true_measure = SciPyWrapper(
        ...     sampler = DigitalNetB2(3,seed=7),
        ...     scipy_distribs = [
        ...         scipy.stats.uniform(loc=1,scale=2),
        ...         scipy.stats.norm(loc=3,scale=4),
        ...         scipy.stats.gamma(a=5,loc=6,scale=7)])
        >>> true_measure.range
        array([[  1.,   3.],
               [-inf,  inf],
               [  6.,  inf]])
        >>> true_measure(4)
        array([[ 2.26535046,  6.36077755, 46.10334984],
               [ 1.37949875,  0.8419074 , 38.66873073],
               [ 2.79562889, -0.65019733, 17.63758514],
               [ 1.91172032,  4.67357136, 55.20163754]])
        
        >>> true_measure = SciPyWrapper(sampler=DigitalNetB2(2,seed=7),scipy_distribs=scipy.stats.beta(a=5,b=1))
        >>> true_measure(4)
        array([[0.93683292, 0.98238098],
               [0.69611329, 0.84454476],
               [0.99733838, 0.50958933],
               [0.84451252, 0.89011392]])
        
        With independent replications 

        >>> x = SciPyWrapper(
        ...     sampler = DigitalNetB2(3,seed=7,replications=2),
        ...     scipy_distribs = [
        ...         scipy.stats.uniform(loc=1,scale=2),
        ...         scipy.stats.norm(loc=3,scale=4),
        ...         scipy.stats.gamma(a=5,loc=6,scale=7)])(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[ 1.49306553, -0.62826013, 49.7677589 ],
                [ 2.36305807,  4.6683678 , 36.07674167],
                [ 1.96279709,  6.34058526, 21.99536017],
                [ 2.8308265 ,  0.84704612, 51.41443437]],
        <BLANKLINE>
               [[ 1.89753782,  7.30327854, 38.9039967 ],
                [ 2.07271848, -3.84426539, 32.72574799],
                [ 1.46428286,  0.81928225, 21.13131114],
                [ 2.50591431,  4.03840677, 51.08541774]]])
    """
    
    def __init__(self, sampler, scipy_distribs):
        r"""
        Args:
            sampler (AbstractDiscreteDistribution): A discrete distribution from which to transform samples.
            scipy_distribs (list): instantiated *continuous univariate* distributions from [`scipy.stats`](https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions).
        """
        self.domain = np.array([[0,1]])
        if not isinstance(sampler,AbstractDiscreteDistribution):
            raise ParameterError("SciPyWrapper requires sampler be an AbstractDiscreteDistribution.")
        self._parse_sampler(sampler)
        self.scipy_distrib = list(scipy_distribs) if not isinstance(scipy_distribs,scipy.stats._distn_infrastructure.rv_continuous_frozen) else [scipy_distribs]
        for sd in self.scipy_distrib:
            if isinstance(sd,scipy.stats._distn_infrastructure.rv_continuous_frozen): continue
            raise ParameterError('''
                SciPyWrapper requires each value of scipy_distribs to be a 
                1 dimensional scipy.stats continuous distribution, 
                see https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions.''')
        self.sds = self.scipy_distrib if len(self.scipy_distrib)>1 else self.scipy_distrib*sampler.d
        if len(self.sds)!=sampler.d:
            raise DimensionError("length of scipy_distribs must match the dimension of the sampler")
        self.range = np.array([sd.interval(1) for sd in self.sds])
        super(SciPyWrapper,self).__init__()
        assert len(self.sds)==self.d and all(isinstance(sdsi,scipy.stats._distn_infrastructure.rv_continuous_frozen) for sdsi in self.sds)

    def _transform(self, x):
        t = np.empty_like(x) 
        for j in range(self.d):
            t[...,j] = self.sds[j].ppf(x[...,j])
        return t
    
    def _weight(self, x):
        rho = np.empty_like(x)
        for j in range(self.d):
            rho[...,j] = self.sds[j].pdf(x[...,j])
        return np.prod(rho,-1)
    
    def _spawn(self, sampler, dimension):
        return SciPyWrapper(sampler,self.scipy_distrib)
    



from .abstract_integrand import AbstractIntegrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Gaussian
import numpy as np
from scipy.special import gamma


class Keister(AbstractIntegrand):
    r"""
    Keister function from [1]. 

    $$f(\boldsymbol{t}) = \pi^{d/2} \cos(\lVert \boldsymbol{t} \rVert_2) \qquad \boldsymbol{T} \sim \mathcal{N}(\boldsymbol{0},\mathsf{I}/2).$$

    Examples:
        >>> integrand = Keister(DigitalNetB2(2,seed=7))
        >>> y = integrand(2**10)
        >>> print("%.4f"%y.mean())
        1.8080
        >>> integrand.true_measure
        Gaussian (AbstractTrueMeasure)
            mean            0
            covariance      2^(-1)
            decomp_type     PCA

        With independent replications

        >>> integrand = Keister(DigitalNetB2(2,seed=7,replications=2**4))
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        1.8024

    **References:**

    1.  B. D. Keister.  
        Multidimensional Quadrature Algorithms.  
        Computers in Physics, 10, pp. 119-122, 1996.
    """

    def __init__(self, sampler):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
        """
        self.sampler = sampler
        self.true_measure = Gaussian(self.sampler,mean=0,covariance=1/2)
        super(Keister,self).__init__(dimension_indv=(),dimension_comb=(),parallel=False)
    
    def g(self, t):
        d = t.shape[-1]
        norm = np.linalg.norm(t,axis=-1)
        k = np.pi**(d/2)*np.cos(norm)
        return k
    
    def _spawn(self, level, sampler):
        return Keister(sampler=sampler)

    @classmethod
    def get_exact_value(self, d):
        """
        Compute the exact analytic value of the Keister integral with dimension $d$. 

        Args:
            d (int): Dimension. 
        
        Returns: 
            mean (float): Exact value of the integral. 
        """
        cosinteg = np.zeros(shape=(d))
        cosinteg[0] = np.sqrt(np.pi) / (2 * np.exp(1 / 4))
        sininteg = np.zeros(shape=(d))
        sininteg[0] = 4.244363835020225e-01
        cosinteg[1] = (1 - sininteg[0]) / 2
        sininteg[1] = cosinteg[0] / 2
        for j in range(2, d):
            cosinteg[j] = ((j-1)*cosinteg[j-2]-sininteg[j-1])/2
            sininteg[j] = ((j-1)*sininteg[j-2]+cosinteg[j-1])/2
        I = (2*(np.pi**(d/2))/gamma(d/2))*cosinteg[d-1]
        return I
    
    def exact_integ(self, *args, **kwargs):
        return self.get_exact_value(*args,**kwargs)




from .common_kernels import (
    KernelGaussian,
    KernelSquaredExponential,
    KernelRationalQuadratic,
    KernelMatern12,
    KernelMatern32,
    KernelMatern52,
)
from .si_dsi_kernels import KernelShiftInvar,KernelShiftInvarCombined,KernelDigShiftInvar,KernelDigShiftInvarAdaptiveAlpha,KernelDigShiftInvarCombined
from .multitask_kernel import KernelMultiTask,KernelMultiTaskDerivs

KernelSI = KernelShiftInvar
KernelSICombined = KernelShiftInvarCombined
KernelDSI = KernelDigShiftInvar
KernelDSIAA = KernelDigShiftInvarAdaptiveAlpha
KernelDSICombined = KernelDigShiftInvarCombined


import torch 
import numpy as np
import itertools 


def fftbr_torch(x):
    r"""
    Torch implementation of the 1 dimensional Bit-Reversed-Order (BRO) Fast Fourier Transform (FFT) along the last dimension. 
    Requires the last dimension of x is already in BRO, so we can skip the first step of the decimation-in-time FFT. 
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = torch.from_numpy(rng.random(8)+1j*rng.random(8)).to(torch.complex64).requires_grad_()
        >>> y = fftbr_torch(x)
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     y.detach()
        tensor([ 1.07+1.14j, -0.32+0.26j, -0.27+0.22j, -0.27-0.34j,  0.10+0.16j,
                -0.15+0.17j,  0.50+0.24j,  0.06-0.02j])
        >>> v = torch.abs(torch.sum(y**2))
        >>> dvdx = torch.autograd.grad(v,x)[0]
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     x.detach()
        tensor([0.25+0.65j, 0.73+0.60j, 0.06+0.20j, 0.61+0.39j, 0.45+0.06j, 0.26+0.09j,
                0.35+0.39j, 0.32+0.85j])
        >>> print("%.4f"%v.detach())
        2.5584
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     dvdx.detach()
        tensor([1.30+0.49j, 1.21+1.45j, 0.79+1.22j, 0.41+0.11j, 1.71+0.61j, 0.79+0.69j,
                0.19+0.51j, 0.12+0.90j])
        >>> fftbr_torch(torch.rand((2,3,4,5,8))+1j*torch.rand((2,3,4,5,8))).shape
        torch.Size([2, 3, 4, 5, 8])
    
    Args:
        x (torch.Tensor): Array of samples at which to run BRO-FFT.
    
    Returns:
        y (torch.Tensor): BRO-FFT values.
    """
    n = x.size(-1)
    assert n&(n-1)==0 # require n is a power of 2
    m = int(np.log2(n))
    shape = list(x.shape)
    ndim = x.ndim
    twos = [2]*m
    xs = x.reshape(shape[:-1]+twos)
    pdims = tuple(itertools.chain(range(ndim-1),range(m+ndim-2,ndim-2,-1)))#[i for i in range(ndim-1)]+[i+ndim-1 for i in range(m-1,-1,-1)]
    xrf = torch.permute(xs,pdims)
    xr = xrf.contiguous().view(shape)
    return torch.fft.fft(xr,norm="ortho")

def ifftbr_torch(x):
    r"""
    Torch implementation of the 1 dimensional Bit-Reversed-Order (BRO) Inverse Fast Fourier Transform (IFFT) along the last dimension.  
    Outputs an array in bit-reversed order, so we can skip the last step of the decimation-in-time IFFT.  
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = torch.from_numpy(rng.random(8)+1j*rng.random(8)).to(torch.complex64).requires_grad_()
        >>> y = ifftbr_torch(x)
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     y.detach()
        tensor([ 1.07+1.14j, -0.29-0.22j,  0.30+0.06j, -0.09+0.02j,  0.03+0.54j,
                -0.04-0.33j, -0.19+0.26j, -0.08+0.36j])
        >>> v = torch.abs(torch.sum(y**2))
        >>> dvdx = torch.autograd.grad(v,x)[0]
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     x.detach()
        tensor([0.25+0.65j, 0.73+0.60j, 0.06+0.20j, 0.61+0.39j, 0.45+0.06j, 0.26+0.09j,
                0.35+0.39j, 0.32+0.85j])
        >>> print("%.4f"%v.detach())
        2.5656
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     dvdx.detach()
        tensor([ 1.15+0.79j,  1.51+1.01j,  0.60+0.86j,  0.06+0.54j, -0.10+0.90j,  0.47+1.37j,
                 0.37+0.20j,  0.83+1.70j])
        >>> ifftbr_torch(torch.rand((2,3,4,5,8))+1j*torch.rand((2,3,4,5,8))).shape
        torch.Size([2, 3, 4, 5, 8])

    Args:
        x (torch.Tensor): Array of samples at which to run BRO-IFFT.
    
    Returns:
        y (torch.Tensor): BRO-IFFT values.
    """
    n = x.size(-1)
    assert n&(n-1)==0 # require n is a power of 2
    m = int(np.log2(n))
    shape = list(x.shape)
    ndim = x.ndim
    twos = [2]*m
    x = torch.fft.ifft(x,norm="ortho")
    xs = x.reshape(shape[:-1]+twos)
    pdims = tuple(itertools.chain(range(ndim-1),range(m+ndim-2,ndim-2,-1)))
    xrf = torch.permute(xs,pdims)
    xr = xrf.contiguous().view(shape)
    return xr

def _fwht_torch(x):
    y = x.clone()+0.
    n = x.size(-1)
    if n<=1: return y
    assert n&(n-1)==0 # require n is a power of 2
    m = int(np.log2(n))
    it = torch.arange(n,dtype=torch.int64,device=x.device).reshape([2]*m) # 2 x 2 x ... x 2 array (size 2^m)
    idx0 = [slice(None)]*(m-1)+[0]
    idx1 = [slice(None)]*(m-1)+[1]
    for k in range(m):
        eps0 = it[tuple(idx0[-(k+1):])].flatten()
        eps1 = it[tuple(idx1[-(k+1):])].flatten()
        y0,y1 = y[(Ellipsis,eps0)],y[(Ellipsis,eps1)]
        y[(Ellipsis,eps0)],y[(Ellipsis,eps1)] = (y0+y1)/np.sqrt(2),(y0-y1)/np.sqrt(2)
    return y
class _FWHTB2Ortho(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        return _fwht_torch(x)
    @staticmethod
    def backward(ctx, dx):
        return _fwht_torch(dx)
def fwht_torch(x):
    r"""
    Torch implementation of the 1 dimensional Fast Walsh Hadamard Transform (FWHT) along the last dimension.  
    Requires the size of the last dimension is a power of 2. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = torch.from_numpy(rng.random(8)).float().requires_grad_()
        >>> y = fwht_torch(x)
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     y.detach()
        tensor([ 1.07, -0.29,  0.12,  0.08,  0.10, -0.45,  0.10, -0.03])
        >>> v = torch.sum(y**2)
        >>> dvdx = torch.autograd.grad(v,x)[0]
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     x.detach()
        tensor([0.25, 0.73, 0.06, 0.61, 0.45, 0.26, 0.35, 0.32])
        >>> print("%.4f"%v.detach())
        1.4694
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     dvdx.detach()
        tensor([0.50, 1.47, 0.11, 1.23, 0.90, 0.51, 0.70, 0.64])
        >>> fwht_torch(torch.rand((2,3,4,5,8))).shape
        torch.Size([2, 3, 4, 5, 8])
    
    Args:
        x (torch.Tensor): Array of samples at which to run FWHT.
    
    Returns:
        y (torch.Tensor): FWHT values.
    """
    return _FWHTB2Ortho.apply(x)

def omega_fwht_torch(m, device=None):
    r"""
    Torch implementation useful when efficiently updating FWHT values after doubling the sample size.  

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> m = 3
        >>> x1 = torch.from_numpy(rng.random((3,5,7,2**m)))
        >>> x2 = torch.from_numpy(rng.random((3,5,7,2**m)))
        >>> x = torch.cat([x1,x2],axis=-1)
        >>> ytrue = fwht_torch(x)
        >>> omega = omega_fwht_torch(m)
        >>> y1 = fwht_torch(x1) 
        >>> y2 = fwht_torch(x2) 
        >>> y = torch.cat([y1+omega*y2,y1-omega*y2],axis=-1)/np.sqrt(2)
        >>> np.allclose(y,ytrue)
        True
    
    Args:
        m (int): Size $2^m$ output. 
    
    Returns:
        y (np.ndarray): $\left(1\right)_{k=0}^{2^m}$. 
    """
    if device is None: device = "cpu"
    return torch.ones(2**m,device=device)

def omega_fftbr_torch(m, device=None):
    r"""
    Torch implementation useful when efficiently updating FFT values after doubling the sample size. 

    Examples:
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> m = 3
        >>> x1 = torch.from_numpy(rng.random((3,5,7,2**m))+1j*rng.random((3,5,7,2**m)))
        >>> x2 = torch.from_numpy(rng.random((3,5,7,2**m))+1j*rng.random((3,5,7,2**m)))
        >>> x = torch.cat([x1,x2],axis=-1)
        >>> ytrue = fftbr_torch(x)
        >>> omega = omega_fftbr_torch(m)
        >>> y1 = fftbr_torch(x1) 
        >>> y2 = fftbr_torch(x2) 
        >>> y = torch.cat([y1+omega*y2,y1-omega*y2],axis=-1)/np.sqrt(2)
        >>> np.allclose(y,ytrue)
        True
    
    Args:
        m (int): Size $2^m$ output. 
    
    Returns:
        y (np.ndarray): $\left(e^{- \pi \mathrm{i} k / 2^m}\right)_{k=0}^{2^m}$. 
    """
    if device is None: device = "cpu"
    return torch.exp(-torch.pi*1j*torch.arange(2**m,device=device)/2**m)



from .abstract_cub_mlqmc import AbstractCubMLQMC
from ..util.data import Data
from ..discrete_distribution import DigitalNetB2,Lattice,Halton
from ..discrete_distribution.abstract_discrete_distribution import AbstractLDDiscreteDistribution
from ..true_measure import Gaussian
from ..integrand import FinancialOption
from ..util import MaxSamplesWarning, ParameterError, MaxLevelsWarning
import numpy as np
from scipy.stats import norm
from time import time
import warnings


class CubMLQMCCont(AbstractCubMLQMC):
    """
    Multilevel Quasi-Monte Carlo stopping criterion with continuation.

    Examples:
        >>> fo = FinancialOption(DigitalNetB2(seed=7,replications=32))
        >>> sc = CubMLQMCCont(fo,abs_tol=1e-3)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.784
            n_total         4718592
            levels          2^(2)
            n_level         [65536 32768 32768 16384]
            mean_level      [1.718 0.051 0.012 0.003]
            var_level       [1.169e-08 2.569e-08 1.850e-08 5.209e-08]
            bias_estimate   2.78e-04
            time_integrate  ...
        CubMLQMCCont (AbstractStoppingCriterion)
            rmse_tol        3.88e-04
            n_init          2^(8)
            n_limit         10000000000
            replications    2^(5)
            levels_min      2^(1)
            levels_max      10
            n_tols          10
            inflate         1.668
            theta_init      2^(-1)
            theta           2^(-3)
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        1
            drift           0
            mean            0
            covariance      1
            decomp_type     PCA
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               1
            replications    2^(5)
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7

    **References:**
        
    1.  [https://github.com/PieterjanRobbe/MultilevelEstimators.jl](https://github.com/PieterjanRobbe/MultilevelEstimators.jl).
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = .05, 
                 rmse_tol = None,
                 n_init = 256,
                 n_limit = 1e10,
                 inflate = 100**(1/9),
                 alpha = .01,   
                 levels_min = 2, 
                 levels_max = 10, 
                 n_tols = 10, 
                 theta_init = 0.5,
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rmse_tol (np.ndarray): Root mean squared error tolerance. 
                If supplied, then absolute tolerance and alpha are ignored in favor of the rmse tolerance. 
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            inflate (float): Coarser tolerance multiplication factor $\geq 1$.
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            levels_min (int): Minimum level of refinement $\geq 2$.
            levels_max (int): Maximum level of refinement $\geq$ `levels_min`.
            n_tols (int): Number of coarser tolerances to run.
            theta_init (float): Initial error splitting constant.
        """
        self.parameters = ['rmse_tol','n_init','n_limit','replications','levels_min',
            'levels_max','n_tols','inflate','theta_init','theta']
        # initialization
        if rmse_tol:
            self.target_tol = float(rmse_tol)
        else: # use absolute tolerance
            self.target_tol =  float(abs_tol) / norm.ppf(1-alpha/2)
        self.n_init = n_init
        self.n_limit = n_limit
        self.levels_min = levels_min
        self.levels_max = levels_max
        self.theta_init = theta_init
        self.theta = theta_init
        self.n_tols = n_tols
        self.alpha = alpha
        self.inflate = inflate
        assert self.inflate>=1
        assert 0<self.alpha<1
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        super(CubMLQMCCont,self).__init__(allowed_distribs=[AbstractLDDiscreteDistribution],allow_vectorized_integrals=False)
        self.replications = self.discrete_distrib.replications 
        assert self.replications>=4, "require at least 4 replications"

    def integrate(self):
        t_start = time()
        data = self._construct_data()
        # Loop over coarser tolerances
        for t in range(self.n_tols):
            self.rmse_tol = self.inflate**(self.n_tols-t-1)*self.target_tol # Set new target tolerance
            self._integrate(data)
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data

    def _construct_data(self):
        data = Data(parameters=[
            'solution',
            'n_total',
            'levels',
            'n_level',
            'mean_level',
            'var_level', 
            'bias_estimate'])
        data.levels = int(self.levels_min+1)
        data.n_level = np.zeros(data.levels,dtype=int)
        data.eval_level = np.ones(data.levels,dtype=bool)
        data.mean_level_reps = np.zeros((data.levels,int(self.replications)))
        data.mean_level = np.tile(0.,data.levels)
        data.var_level = np.tile(np.inf,data.levels)
        data.cost_level = np.tile(0.,data.levels)
        data.var_cost_ratio_level = np.tile(np.inf,data.levels)
        data.bias_estimate = np.inf
        data.level_integrands = []
        return data 
    
    def _integrate(self, data):
        #self.theta = self.theta_init
        data.levels = int(self.levels_min+1)

        converged = False
        while not converged:
            # Ensure that we have samples on the finest level
            self.update_data(data)
            self._update_theta(data)

            while self._varest(data) > (1-self.theta)*self.rmse_tol**2:
                efficient_level = np.argmax(data.var_cost_ratio_level[:data.levels])
                data.eval_level[efficient_level] = True

                # Check if over sample budget
                total_next_samples = (self.replications*data.eval_level*data.n_level*2).sum()
                if (data.n_total + total_next_samples) > self.n_limit:
                    warning_s = """
                    Already generated %d samples.
                    Trying to generate %d new samples, which would exceed n_limit = %d.
                    Stopping integration process.
                    Note that error tolerances may no longer be satisfied""" \
                    % (int(data.n_total), int(total_next_samples), int(self.n_limit))
                    warnings.warn(warning_s, MaxSamplesWarning)
                    return

                self.update_data(data)
                self._update_theta(data)

            # Check for convergence
            converged = self._rmse(data) < self.rmse_tol
            if not converged:
                if data.levels == self.levels_max:
                    warnings.warn(
                        'Failed to achieve weak convergence. levels == levels_max.',
                        MaxLevelsWarning)
                    converged = True
                else:
                    self._add_level(data)

    def _update_theta(self, data):
        # Update error splitting parameter
        max_levels = len(data.n_level)
        A = np.ones((2,2))
        A[:,0] = range(max_levels-2, max_levels)
        y = np.ones(2)
        y[0] = np.log2(abs(data.mean_level_reps[max_levels-2].mean()))
        y[1] = np.log2(abs(data.mean_level_reps[max_levels-1].mean()))
        x = np.linalg.lstsq(A, y, rcond=None)[0]
        alpha = max(.5,-x[0])
        real_bias = 2**(x[1]+max_levels*x[0]) / (2**alpha - 1)
        self.theta = max(0.01, min(0.125, (real_bias/self.rmse_tol)**2))

    def _rmse(self, data):
        # Returns an estimate for the root mean square error
        return np.sqrt(self._mse(data))

    def _mse(self, data):
        # Returns an estimate for the mean square error
        return (1-self.theta)*self._varest(data) + self.theta*data.bias_estimate**2

    def _varest(self, data):
        # Returns the variance of the estimator
        return data.var_level[:data.levels].sum()



from .abstract_true_measure import AbstractTrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution import DigitalNetB2
import numpy as np
from scipy.stats import norm
from typing import Union


class JohnsonsSU(AbstractTrueMeasure):
    r"""
    Johnson's $S_U$-distribution with independent marginals as described in [https://en.wikipedia.org/wiki/Johnson%27s_SU-distribution](https://en.wikipedia.org/wiki/Johnson%27s_SU-distribution).

    Examples:
        >>> true_measure = JohnsonsSU(DigitalNetB2(2,seed=7),gamma=1,xi=2,delta=3,lam=4)
        >>> true_measure(4)
        array([[ 1.44849599,  2.49715741],
               [-0.83646172,  0.38970902],
               [ 3.67068094, -2.33911196],
               [ 0.38940887,  0.84843818]])
        >>> true_measure
        JohnsonsSU (AbstractTrueMeasure)
            gamma           1
            xi              2^(1)
            delta           3
            lam             2^(2)
        
        With independent replications 

        >>> x = JohnsonsSU(DigitalNetB2(3,seed=7,replications=2),gamma=1,xi=2,delta=3,lam=4)(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[-0.36735335, -0.71750135,  1.55387818],
                [ 1.29233112,  1.21788962,  0.3870404 ],
                [ 0.57599197,  1.78008445, -1.53756327],
                [ 2.50112084, -0.14204369,  1.67333839]],
        <BLANKLINE>
               [[ 0.45920696,  2.10110361,  0.66122546],
                [ 0.76973983, -2.12724026,  0.02868419],
                [-0.43948474, -0.1525475 , -1.71041918],
                [ 1.57765245,  1.00275   ,  1.64972468]]])
    """

    def __init__(self, sampler, gamma=1, xi=1, delta=2, lam=2):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            gamma (Union[float,np.ndarray]): First parameter $\gamma$.
            xi (Union[float,np.ndarray]): Second parameter $\xi$.
            delta (Union[float,np.ndarray]): Third parameter $\delta > 0$. 
            lam (Union[float,np.ndarray]): Fourth parameter $\lambda > 0$.
        """
        self.parameters = ['gamma', 'xi', 'delta', 'lam']
        self.domain = np.array([[0,1]])
        self.range = np.array([[-np.inf,np.inf]])
        self._parse_sampler(sampler)
        self.gamma = gamma
        self.xi = xi
        self.delta = delta
        self.lam = lam
        self._gamma = np.array(gamma)
        if self._gamma.size==1:
            self._gamma = self._gamma.item()*np.ones(self.d)
        self._xi = np.array(xi)
        if self._xi.size==1:
            self._xi = self._xi.item()*np.ones(self.d)
        self._delta = np.array(delta)
        if self._delta.size==1:
            self._delta = self._delta.item()*np.ones(self.d)
        self._lam = np.array(lam)
        if self._lam.size==1:
            self._lam = self._lam.item()*np.ones(self.d)
        if not (self._gamma.shape==(self.d,) and self._xi.shape==(self.d,) and self._delta.shape==(self.d,) and self._lam.shape==(self.d,)):
            raise DimensionError("all Johnson's S_U parameters be scalar or have length equal to dimension.")
        if not ((self._delta>0).all() and (self._lam>0).all()):
            raise ParameterError("delta and lam must be all be positive")
        super(JohnsonsSU,self).__init__()
        assert self._gamma.shape==(self.d,) and self._xi.shape==(self.d,) and self._delta.shape==(self.d,) and self._lam.shape==(self.d,)

    def _transform(self, x):
        return self._lam*np.sinh((norm.ppf(x)-self._gamma)/self._delta)+self._xi
    
    def _weight(self, x):
        term1 = (x-self._xi)/self._lam
        term2 = self._delta/(self._lam*np.sqrt(2*np.pi)) * 1/np.sqrt(1+term1**2)
        term3 = np.exp(-1/2*(self._gamma+self._delta*np.arcsinh(term1))**2)
        return np.prod(term2*term3,-1)
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = JohnsonsSU(sampler,gamma=self._gamma,xi=self._xi,delta=self._delta,lam=self._lam)
        else:
            gamma = self._gamma[0]
            xi = self._xi[0]
            delta = self._delta[0]
            lam = self._lam[0]
            if not ( all(self._gamma==gamma) and all(self._xi==xi) and all(self._delta==delta) and all(self._lam==lam) ):
                raise DimensionError('''
                    In order to spawn a JohnsonsSU measure
                    gamma must all be the same and 
                    xi must all be the same and 
                    delta must all be the same and 
                    lam (lambda) must all be the same.''')
            spawn = JohnsonsSU(sampler,gamma=gamma,xi=xi,delta=delta,lam=lam)
        return spawn



import qmcpy as qp 
import numpy as np

def mlmc_test(
    integrand,
    n = 20000,
    l = 8,
    n_init = 200,
    rmse_tols = np.array([.005, 0.01, 0.02, 0.05, 0.1]),
    levels_min = 2,
    levels_max = 10,
    ):
    r"""
    Multilevel Monte Carlo test routine.

    Examples:
        >>> fo = qp.FinancialOption(
        ...     sampler=qp.IIDStdUniform(seed=7),
        ...     option = "ASIAN",
        ...     asian_mean = "GEOMETRIC",
        ...     volatility = 0.2, 
        ...     start_price = 100, 
        ...     strike_price = 100, 
        ...     interest_rate = 0.05, 
        ...     t_final = 1)
        >>> print('Exact Value: %s'%fo.get_exact_value_inf_dim())
        Exact Value: 5.546818633789201
        >>> mlmc_test(fo)
        Convergence tests, kurtosis, telescoping sum check using N =  20000 samples
            l              ave(Pf-Pc)     ave(Pf)        var(Pf-Pc)     var(Pf)        kurtosis       check          cost
            0              5.4486e+00     5.4486e+00     5.673e+01      5.673e+01      0.00e+00       0.00e+00       2.00e+00
            1              1.4925e-01     5.5937e+00     3.839e+00      5.838e+01      5.48e+00       1.14e-02       4.00e+00
            2              3.5921e-02     5.6024e+00     9.585e-01      5.990e+01      5.59e+00       7.86e-02       8.00e+00
            3              8.7217e-03     5.5128e+00     2.332e-01      5.828e+01      5.36e+00       2.92e-01       1.60e+01
            4              1.9773e-03     5.6850e+00     6.021e-02      6.081e+01      5.46e+00       5.12e-01       3.20e+01
            5              9.5925e-04     5.5628e+00     1.512e-02      5.939e+01      5.37e+00       3.71e-01       6.40e+01
            6              8.5998e-04     5.5706e+00     3.773e-03      5.995e+01      5.48e+00       2.10e-02       1.28e+02
            7              1.3592e-04     5.4359e+00     9.285e-04      5.808e+01      5.51e+00       4.13e-01       2.56e+02
            8              3.4520e-05     5.5322e+00     2.313e-04      5.881e+01      5.57e+00       2.96e-01       5.12e+02
        Linear regression estimates of MLMC parameters
            alpha = 1.617207  (exponent for MLMC weak convergence)
            beta  = 2.000355  (exponent for MLMC variance)
            gamma = 1.000000  (exponent for MLMC cost)
        MLMC complexity tests
            rmse_tol       value          mlmc_cost      std_cost       savings        N_l
            5.000e-03      5.545e+00      3.339e+07      1.038e+08      3.11           8605392      1566846      559701       198886       70359        
            1.000e-02      5.539e+00      7.272e+06      1.243e+07      1.71           2009192      365451       130781       46623        
            2.000e-02      5.549e+00      1.827e+06      3.108e+06      1.70           503397       91821        33196        11736        
            5.000e-02      5.474e+00      2.324e+05      2.556e+05      1.10           71432        13143        4617         
            1.000e-01      5.466e+00      6.220e+04      6.389e+04      1.03           19477        3361         1225         
        
    Args:
        integrand (AbstractIntegrand): multilevel integrand
        n (int): number of samples for convergence tests
        l (int): number of levels for convergence tests
        n_init (int): initial number of samples for MLMC calcs
        rmse_tols (np.ndarray): desired accuracy array for MLMC calcs
        levels_min (int): minimum number of levels for MLMC calcs
        levels_max (int): maximum number of levels for MLMC calcs
    """
    # first, convergence tests
    n = 100*np.ceil(n/100) # make N a multiple of 100
    print('Convergence tests, kurtosis, telescoping sum check using N =%7d samples'%n)
    print('    %-15s%-15s%-15s%-15s%-15s%-15s%-15s%s'\
        %('l','ave(Pf-Pc)','ave(Pf)','var(Pf-Pc)','var(Pf)','kurtosis','check','cost'))
    del1 = np.array([])
    del2 = np.array([])
    var1 = np.array([])
    var2 = np.array([])
    kur1 = np.array([])
    chk1 = np.array([])
    cost = np.array([])
    integrand_spawns = integrand.spawn(levels=np.arange(l+1))
    for ll in range(l+1):
        sums = np.zeros(6)
        cst = 0
        integrand_spawn = integrand_spawns[ll]
        for j in range(1,101):
            # evaluate integral at sampleing points samples
            samples = integrand_spawn.discrete_distrib.gen_samples(n=n/100)
            Pc,Pf = integrand_spawn.f(samples)
            dP = Pf-Pc
            sums_j = np.array([
                np.sum(dP),
                np.sum(dP**2),
                np.sum(dP**3),
                np.sum(dP**4),
                np.sum(Pf),
                np.sum(Pf**2),
            ])
            cst_j = integrand_spawn.cost*(n/100)
            sums = sums + sums_j/n
            cst = cst + cst_j/n
        if ll == 0:
            kurt = 0.
        else:
            kurt = ( sums[3] - 4*sums[2]*sums[0] + 6*sums[1]*sums[0]**2 - 
                     3*sums[0]*sums[0]**3 ) /  (sums[1]-sums[0]**2)**2.
        cost = np.hstack((cost, cst))
        del1 = np.hstack((del1, sums[0]))
        del2 = np.hstack((del2, sums[4]))
        var1 = np.hstack((var1, sums[1]-sums[0]**2))
        var2 = np.hstack((var2, sums[5]-sums[4]**2))
        var2 = np.maximum(var2, 1e-10) # fix for cases with var=0
        kur1 = np.hstack((kur1, kurt))
        if ll == 0:
            check = 0
        else:
            check = abs( del1[ll] + del2[ll-1] - del2[ll]) / \
                    ( 3.*( np.sqrt(var1[ll]) + np.sqrt(var2[ll-1]) + np.sqrt(var2[ll]) ) / np.sqrt(n))
        chk1 = np.hstack((chk1, check))

        print('    %-15d%-15.4e%-15.4e%-15.3e%-15.3e%-15.2e%-15.2e%.2e'\
              %(ll,del1[ll],del2[ll],var1[ll],var2[ll],kur1[ll],chk1[ll],cst))
    # print out a warning if kurtosis or consistency check looks bad
    if kur1[-1] > 100.:
        print('WARNING: kurtosis on finest level = %f'%kur1[-1])
        print(' indicates MLMC correction dominated by a few rare paths;')
        print(' for information on the connection to variance of sample variances,')
        print(' see http://mathworld.wolfram.com/SampleVarianceDistribution.html\n')
    if np.max(chk1) > 1.:
        print('WARNING: maximum consistency error = %f'%max(chk1))
        print(' indicates identity E[Pf-Pc] = E[Pf] - E[Pc] not satisfied;')
        print(' to be more certain, re-run mlmc_test with larger N\n')
    # use linear regression to estimate alpha, beta and gamma
    l1 = 2
    l2 = l+1
    x = np.ones((l2+1-l1,2))
    x[:,1] = np.arange(l1,l2+1)
    pa = np.linalg.lstsq(x,np.log2(np.absolute(del1[(l1-1):l2])),rcond=None)[0]
    alpha = -pa[1]
    pb = np.linalg.lstsq(x,np.log2(np.absolute(var1[(l1-1):l2])),rcond=None)[0]
    beta = -pb[1]
    pg = np.linalg.lstsq(x,np.log2(np.absolute(cost[(l1-1):l2])),rcond=None)[0]
    gamma = pg[1]
    print('Linear regression estimates of MLMC parameters')
    print('    alpha = %f  (exponent for MLMC weak convergence)'%alpha)
    print('    beta  = %f  (exponent for MLMC variance)'%beta)
    print('    gamma = %f  (exponent for MLMC cost)'%gamma)
    #second, mlmc complexity tests
    print('MLMC complexity tests')
    print('    %-15s%-15s%-15s%-15s%-15s%s'\
        %('rmse_tol','value','mlmc_cost','std_cost','savings','N_l'))
    alpha = np.maximum(alpha,0.5)
    beta  = np.maximum(beta,0.5)
    theta = 0.25
    for i in range(len(rmse_tols)):
        mlmc_qmcpy = qp.CubMLMC(integrand,
            rmse_tol = rmse_tols[i],
            n_init = n_init,
            levels_min = levels_min,
            levels_max = levels_max,
            alpha0 = alpha,
            beta0 = beta,
            gamma0 = gamma)
        sol,data = mlmc_qmcpy.integrate()
        p = data.solution
        nl = data.n_level
        cl = data.cost_per_sample
        mlmc_cost = sum(nl*cl)
        idx = np.minimum(len(var2),len(nl))-1
        std_cost = var2[idx]*cl[-1] / ((1.-theta)*rmse_tols[i]**2)
        print('    %-15.3e%-15.3e%-15.3e%-15.3e%-15.2f%s'\
            %(rmse_tols[i], p, mlmc_cost, std_cost, std_cost/mlmc_cost,''.join('%-13d'%nli for nli in nl)))



from .abstract_cub_mlqmc import AbstractCubMLQMC
from ..util.data import Data
from ..discrete_distribution import DigitalNetB2,Lattice,Halton
from ..discrete_distribution.abstract_discrete_distribution import AbstractLDDiscreteDistribution
from ..true_measure import Gaussian
from ..integrand import FinancialOption
from ..util import MaxSamplesWarning, ParameterError, MaxLevelsWarning
import numpy as np
from scipy.stats import norm
from time import time
import warnings


class CubMLQMC(AbstractCubMLQMC):
    """
    Multilevel Quasi-Monte Carlo stopping criterion.

    Examples:
        >>> fo = FinancialOption(DigitalNetB2(seed=7,replications=32))
        >>> sc = CubMLQMC(fo,abs_tol=3e-3)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.784
            n_total         2359296
            levels          2^(2)
            n_level         [32768 16384 16384  8192]
            mean_level      [1.718 0.051 0.012 0.003]
            var_level       [7.119e-08 1.409e-07 9.668e-08 1.852e-07]
            bias_estimate   2.99e-04
            time_integrate  ...
        CubMLQMC (AbstractStoppingCriterion)
            rmse_tol        0.001
            n_init          2^(8)
            n_limit         10000000000
            replications    2^(5)
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        1
            drift           0
            mean            0
            covariance      1
            decomp_type     PCA
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               1
            replications    2^(5)
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7

    **References:**
        
    1.  M.B. Giles and B.J. Waterhouse.  
        'Multilevel quasi-Monte Carlo path simulation'.  
        pp.165-181 in Advanced Financial Modelling, in Radon Series on Computational and Applied Mathematics, de Gruyter, 2009.  
        [http://people.maths.ox.ac.uk/~gilesm/files/radon.pdf](http://people.maths.ox.ac.uk/~gilesm/files/radon.pdf).
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = .05, 
                 rmse_tol = None, 
                 n_init = 256, 
                 n_limit = 1e10,
                 alpha = .01, 
                 levels_min = 2,
                 levels_max = 10,
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rmse_tol (np.ndarray): Root mean squared error tolerance. 
                If supplied, then absolute tolerance and alpha are ignored in favor of the rmse tolerance. 
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            levels_min (int): Minimum level of refinement $\geq 2$.
            levels_max (int): Maximum level of refinement $\geq$ `levels_min`.
        """
        self.parameters = ['rmse_tol','n_init','n_limit','replications']
        # initialization
        if rmse_tol:
            self.rmse_tol = float(rmse_tol)
        else: # use absolute tolerance
            self.rmse_tol =  float(abs_tol) / norm.ppf(1-alpha/2)
        self.alpha = alpha 
        assert 0<self.alpha<1
        self.n_init = n_init
        self.n_limit = n_limit
        self.levels_min = levels_min
        self.levels_max = levels_max
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        super(CubMLQMC,self).__init__(allowed_distribs=[AbstractLDDiscreteDistribution],allow_vectorized_integrals=False)
        self.replications = self.discrete_distrib.replications 
        assert self.replications>=4, "require at least 4 replications"

    def integrate(self):
        t_start = time()
        data = Data(parameters=[
            'solution',
            'n_total',
            'levels',
            'n_level',
            'mean_level',
            'var_level', 
            'bias_estimate'])
        data.levels = int(self.levels_min+1)
        data.n_level = np.zeros(data.levels,dtype=int)
        data.eval_level = np.ones(data.levels,dtype=bool)
        data.mean_level_reps = np.zeros((data.levels,int(self.replications)))
        data.mean_level = np.tile(0.,data.levels)
        data.var_level = np.tile(np.inf,data.levels)
        data.cost_level = np.tile(0.,data.levels)
        data.var_cost_ratio_level = np.tile(np.inf,data.levels)
        data.bias_estimate = np.inf
        data.level_integrands = []
        while True:
            self.update_data(data)
            if data.var_level.sum() > (self.rmse_tol**2/2.):
                # double N_l on level with largest V_l/(2^l*N_l)
                efficient_level = np.argmax(data.var_cost_ratio_level)
                data.eval_level[efficient_level] = True
            elif data.bias_estimate > (self.rmse_tol/np.sqrt(2.)):
                if data.levels == self.levels_max + 1:
                        warnings.warn("Failed to achieve weak convergence. levels == levels_max.", MaxLevelsWarning)
                # add another level
                self._add_level(data)
            else:
                # both conditions met
                break
            total_next_samples = (self.replications*data.eval_level*data.n_level*2).sum()
            if (data.n_total + total_next_samples) > self.n_limit:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples, which would exceed n_limit = %d.
                Stopping integration process.
                Note that error tolerances may no longer be satisfied""" \
                % (int(data.n_total), int(total_next_samples), int(self.n_limit))
                warnings.warn(warning_s, MaxSamplesWarning)
                break
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data



from .gaussian import Gaussian
from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
from .abstract_true_measure import AbstractTrueMeasure
from ..discrete_distribution import DigitalNetB2
from ..util import DimensionError, ParameterError
import numpy as np
from scipy.special import kv, gamma
from typing import Union


class MaternGP(Gaussian):
    r"""
    A Gaussian process with Matérn covariance kernel.

    Examples:
        >>> true_measure = MaternGP(DigitalNetB2(dimension=3,seed=7),points=np.linspace(0,1,3)[:,None],nu=3/2,length_scale=[3,4,5],variance=0.01,mean=np.array([.3,.4,.5]))
        >>> true_measure(4)
        array([[0.3515401 , 0.43083384, 0.51801119],
               [0.20272448, 0.31312011, 0.4241431 ],
               [0.40189226, 0.53502934, 0.63826677],
               [0.29943567, 0.38491661, 0.48296594]])
        >>> true_measure
        MaternGP (AbstractTrueMeasure)
            mean            [0.3 0.4 0.5]
            covariance      [[0.01  0.01  0.01 ]
                             [0.01  0.01  0.01 ]
                             [0.009 0.01  0.01 ]]
            decomp_type     PCA
        
        With independent replications 

        >>> x = MaternGP(DigitalNetB2(dimension=3,seed=7,replications=2),points=np.linspace(0,1,3)[:,None],nu=3/2,length_scale=[3,4,5],variance=0.01,mean=np.array([.3,.4,.5]))(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[0.21490091, 0.33078241, 0.45151042],
                [0.35465127, 0.44705898, 0.53793358],
                [0.31091595, 0.39868187, 0.47660193],
                [0.42419919, 0.53572415, 0.64674883]],
        <BLANKLINE>
               [[0.31010701, 0.38522001, 0.46670381],
                [0.27221177, 0.413546  , 0.54101758],
                [0.2147053 , 0.33293508, 0.43572791],
                [0.37343973, 0.46534628, 0.56356714]]])
    
    **References:**

    1.  [`sklearn.gaussian_process.kernels.Matern`](https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.MaternGP.html).
    
    2.  [https://en.wikipedia.org/wiki/Mat%C3%A9rn_covariance_function](https://en.wikipedia.org/wiki/Mat%C3%A9rn_covariance_function).
    """

    def __init__(self, sampler, points, length_scale=1.0, nu=1.5, variance=1.0, mean=0., nugget=1e-6, decomp_type='PCA'):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            points (np.ndarray): The positions of points on a metric space. The array should have shape $(d,k)$ where $d$ is the dimension of the sampler and $k$ is the latent dimension. 
            nu (float): The "smoothness" of the MaternGP function, e.g.,
                
                - $\nu = 1/2$ is equivalent to the absolute exponential kernel, 
                - $\nu = 3/2$ implies a once-differentiable function,
                - $\nu = 5/2$ implies twice differentiability. 
                - as $\nu \to \infty$ the kernel becomes equivalent to the RBF kernel, see [`sklearn.gaussian_process.kernels.RBF`](https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.RBF.html#sklearn.gaussian_process.kernels.RBF). 
                
                Note that when $\nu \notin \{1/2, 3/2, 5/2, \infty \}$ the kernel is around $10$ times slower to evaluate. 
            length_scale (Union[float,np.ndarray]): Determines "peakiness", or how correlated two points are based on their distance.
            variance (float): Global scaling factor.
            mean (Union[float,np.ndarray]): Mean vectorfor multivariante `Gaussian`. 
            nugget (float): Positive nugget to add to diagonal.
            decomp_type (str): Method for decomposition for covariance matrix. Options include
             
                - `'PCA'` for principal component analysis, or 
                - `'Cholesky'` for cholesky decomposition.
        """
        if not (isinstance(sampler, AbstractDiscreteDistribution) or isinstance(sampler, AbstractTrueMeasure)):
            raise ParameterError("sampler input should either be an AbstractDiscreteDistribution or AbstractTrueMeasure.")
        if not (isinstance(points, np.ndarray) and (points.ndim==1 or points.ndim==2)):
            raise ParameterError("points must be a one or two dimensional np.ndarray.")
        if points.ndim==1: points = points[:,None] 
        assert points.ndim==2 and points.shape[0]==sampler.d, "points should be a two dimenssion array with the number of points equal to the dimension of the sampler"
        mean = np.array(mean)
        if mean.size==1: mean = mean.item()*np.ones(sampler.d)
        assert mean.shape==(sampler.d,), "mean should be a length d vector"
        assert np.isscalar(nu) and nu>0, "nu should be a positive scalar"
        length_scale = np.array(length_scale) 
        if length_scale.size==1: length_scale = length_scale.item()*np.ones(sampler.d)
        assert length_scale.shape==(sampler.d,) and (length_scale>0).all(), "length_scale should be a vector with length equal to the dimension of the sampler"
        assert np.isscalar(variance) and variance>0, "length_scale should be a positive scalar"
        assert np.isscalar(nugget) and nugget>0, "nugget should be a positive scalar"
        self.points = points
        self.length_scale = length_scale
        self.nu = nu
        self.variance = variance
        dists = np.linalg.norm(points[...,:,None,:]-points[...,None,:,:],axis=-1)
        if nu==1/2:
            covariance = np.exp(-dists/self.length_scale)
        elif nu==3/2:
            covariance = (1+np.sqrt(3)*dists/self.length_scale)*np.exp(-np.sqrt(3)*dists/self.length_scale)
        elif nu==5/2:
            covariance = ((1+np.sqrt(5)*dists/self.length_scale+5*dists**2/(3*self.length_scale**2))*np.exp(-np.sqrt(5)*dists/self.length_scale))
        elif nu==np.inf:
            covariance = np.exp(-dists**2/(2*self.length_scale**2))
        else:
            k = np.sqrt(2*nu)*dists/self.length_scale
            covariance = 2**(1-nu)/gamma(nu)*k**nu*kv(nu,k)
        covariance = variance*covariance+nugget*np.eye(sampler.d)
        super().__init__(sampler, mean=mean, covariance=covariance, decomp_type=decomp_type)

    def _spawn(self, sampler):
        return MaternGP(sampler, self.points, length_scale=self.length_scale, nu=self.nu, variance=self.variance, 
                      mean=self.mean, decomp_type=self.decomp_type)
        



from .abstract_integrand import AbstractIntegrand
from .keister import Keister
from .box_integral import BoxIntegral
from ..util import ParameterError
from ..true_measure import Uniform
from ..discrete_distribution import DigitalNetB2
import numpy as np
from itertools import combinations
import scipy.special


class SensitivityIndices(AbstractIntegrand):
    r"""
    Sensitivity indices i.e. normalized Sobol' Indices. 
    
    Examples:
        Singleton indices

        >>> function = Keister(DigitalNetB2(dimension=4,seed=7))
        >>> integrand = SensitivityIndices(function,indices='singletons')
        >>> integrand.indices
        array([[ True, False, False, False],
               [False,  True, False, False],
               [False, False,  True, False],
               [False, False, False,  True]])
        >>> y = integrand(2**10)
        >>> y.shape
        (2, 3, 4, 1024)
        >>> ymean = y.mean(-1)
        >>> ymean.shape
        (2, 3, 4)
        >>> sigma_hat = ymean[:,2,:]-ymean[:,1,:]**2
        >>> sigma_hat.shape 
        (2, 4)
        >>> sigma_hat
        array([[17.81421941, 17.81421941, 17.81421941, 17.81421941],
               [17.81421941, 17.81421941, 17.81421941, 17.81421941]])
        >>> closed_total_approx = ymean[:,0]/sigma_hat
        >>> closed_total_approx.shape 
        (2, 4)
        >>> closed_total_approx
        array([[0.23756849, 0.2423556 , 0.24844127, 0.24771739],
               [0.25090825, 0.25344588, 0.24531633, 0.26244379]])

        Check what all indices look like for $d=3$ 

        >>> integrand = SensitivityIndices(Keister(DigitalNetB2(dimension=3,seed=7)),indices='all')
        >>> integrand.indices
        array([[ True, False, False],
               [False,  True, False],
               [False, False,  True],
               [ True,  True, False],
               [ True, False,  True],
               [False,  True,  True]])

        Vectorized function for all singletons and pairs of dimensions

        >>> function = BoxIntegral(DigitalNetB2(dimension=4,seed=7,replications=2**4),s=np.arange(1,31).reshape((5,6)))
        >>> indices = np.zeros((function.d,function.d,function.d),dtype=bool) 
        >>> r = np.arange(function.d) 
        >>> indices[r,:,r] = True 
        >>> indices[:,r,r] = True 
        >>> integrand = SensitivityIndices(function,indices=indices)
        >>> integrand.indices.shape
        (4, 4, 4)
        >>> integrand.indices
        array([[[ True, False, False, False],
                [ True,  True, False, False],
                [ True, False,  True, False],
                [ True, False, False,  True]],
        <BLANKLINE>
               [[ True,  True, False, False],
                [False,  True, False, False],
                [False,  True,  True, False],
                [False,  True, False,  True]],
        <BLANKLINE>
               [[ True, False,  True, False],
                [False,  True,  True, False],
                [False, False,  True, False],
                [False, False,  True,  True]],
        <BLANKLINE>
               [[ True, False, False,  True],
                [False,  True, False,  True],
                [False, False,  True,  True],
                [False, False, False,  True]]])
        >>> y = integrand(2**10)
        >>> y.shape 
        (2, 3, 4, 4, 5, 6, 16, 1024)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (2, 3, 4, 4, 5, 6, 16)
        >>> muhathat = muhats.mean(-1) 
        >>> muhathat.shape 
        (2, 3, 4, 4, 5, 6)
        >>> sigma_hat = muhathat[:,2,:]-muhathat[:,1,:]**2
        >>> sigma_hat.shape
        (2, 4, 4, 5, 6)
        >>> closed_total_approx = muhathat[:,0]/sigma_hat
        >>> closed_total_approx.shape
        (2, 4, 4, 5, 6)
    
    **References:**
    
    1.  Aleksei G. Sorokin and Jagadeeswaran Rathinavel.  
        On Bounding and Approximating Functions of Multiple Expectations Using Quasi-Monte Carlo.  
        International Conference on Monte Carlo and Quasi-Monte Carlo Methods in Scientific Computing.  
        Cham: Springer International Publishing, 2022.  
        [https://link.springer.com/chapter/10.1007/978-3-031-59762-6_29](https://link.springer.com/chapter/10.1007/978-3-031-59762-6_29). 
    
    2.  Art B. Owen.  
        Monte Carlo theory, methods and examples.  
        Appendix A. Equations (A.16) and (A.18). 2013.
        [https://artowen.su.domains/mc/A-anova.pdf](https://artowen.su.domains/mc/A-anova.pdf).
    """
    def __init__(self, integrand, indices='singletons'):
        r"""
        Args:
            integrand (AbstractIntegrand): Integrand to find sensitivity indices of.
            indices (np.ndarray): Bool array with shape $(\dots,d)$ where each length $d$ vector item indicates which dimensions are active in the subset.
            
                - The default `indices='singletons'` sets `indices=np.eye(d,dtype=bool)`.
                - Setting `incides='all'` sets `indices = np.array([[bool(int(b)) for b in np.binary_repr(i,width=d)] for i in range(1,2**d-1)],dtype=bool)`
        """
        self.parameters = ['indices']
        self.integrand = integrand
        self.dtilde = self.integrand.d
        assert self.dtilde>1, "SensitivityIndices does not make sense for d=1"
        self.indices = indices
        if isinstance(self.indices,str) and self.indices=='singletons':
            self.indices = np.eye(self.dtilde,dtype=bool)
        elif isinstance(self.indices,str) and self.indices=='all':
            self.indices = np.zeros((0,self.dtilde),dtype=bool)
            for r in range(1,self.dtilde):
                idxs_r = np.zeros((int(scipy.special.comb(self.dtilde,r)),self.dtilde),dtype=bool)
                for i,comb in enumerate(combinations(range(self.dtilde),r)):
                    idxs_r[i,comb] = True 
                self.indices = np.vstack([self.indices,idxs_r])
        self.indices = np.atleast_1d(self.indices)
        assert self.indices.dtype==bool and self.indices.ndim>=1 and self.indices.shape[-1]==self.dtilde 
        assert not (self.indices==self.indices[...,0,None]).all(-1).any(), "indices cannot include the emptyset or the set of all dimensions"
        self.not_indices = ~self.indices
        # sensitivity_index
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib.spawn(s=1,dimensions=[2*self.dtilde])[0]
        self.sampler = self.integrand.sampler
        self.i_slice = (slice(None),)*len(self.integrand.d_indv)
        super(SensitivityIndices,self).__init__(
            dimension_indv = (2,3)+self.indices.shape[:-1]+self.integrand.d_indv,
            dimension_comb = (2,)+self.indices.shape[:-1]+self.integrand.d_indv,
            parallel = False)
        self.d = 2*self.dtilde
    
    def f(self, x, *args, **kwargs):
        if 'compute_flags' in kwargs:
            compute_flags = kwargs['compute_flags']
            del kwargs['compute_flags']
        else:
            compute_flags = np.ones(self.d_indv,dtype=bool)
        assert compute_flags.shape==self.d_indv
        z = x[...,self.dtilde:]
        x = x[...,:self.dtilde]
        v = np.zeros_like(x)
        y = np.zeros(self.d_indv+x.shape[:-1],dtype=float)
        f_x = self.integrand.f(x,*args,**kwargs)
        f_z = self.integrand.f(z,*args,**kwargs)
        for i in np.ndindex(self.indices.shape[:-1]):
            flags_closed = compute_flags[(0,0)+i]
            flags_total = compute_flags[(1,0)+i]
            flags_i = flags_closed|flags_total
            if not flags_i.any(): continue
            u_bool = self.indices[i]
            not_u_bool = self.not_indices[i]
            v[...,u_bool] = x[...,u_bool]
            v[...,not_u_bool] = z[...,not_u_bool]
            f_v = self.integrand.f(v,compute_flags=flags_i,*args,**kwargs)
            y[(0,0)+i+self.i_slice] = f_x*(f_v-f_z) # A.18
            y[(1,0)+i+self.i_slice] = (f_z-f_v)**2/2 # A.16
            y[(slice(None),1)+i+self.i_slice] = f_x[(None,)+self.i_slice] # mu
            y[(slice(None),2)+i+self.i_slice] = f_x[(None,)+self.i_slice]**2 # sigma^2+mu^2
            # here we copy mu and sigma^2+mu^2 since if these these were not copied there is a chance the bounds could change 
            # for mu and/or sigma and then an index which was previously approximated sufficiently woulud become insufficientlly approximated 
            # and it would then be difficult ot go back and resample the numerator for that approximation
        return y
    
    def _spawn(self, level, sampler):
        new_integrand = self.integrand.spawn(level,sampler)
        return SensitivityIndices(
            integrand = new_integrand,
            indices = self.indices)
    
    def bound_fun(self, bound_low, bound_high):
        tau_low,mu_low,f2_low = bound_low[:,0],bound_low[:,1],bound_low[:,2]
        tau_high,mu_high,f2_high = bound_high[:,0],bound_high[:,1],bound_high[:,2]
        sigma2_low1,sigma2_low2 = f2_high-mu_low**2,f2_high-mu_high**2
        comb_bounds_low = np.clip(np.minimum.reduce([tau_low/sigma2_low1,tau_low/sigma2_low2]),0,1)
        sigma2_high1,sigma2_high2 = f2_low-mu_low**2,f2_low-mu_high**2
        comb_bounds_high = np.clip(np.maximum.reduce([tau_high/sigma2_high1,tau_high/sigma2_high2]),0,1)
        violated = (sigma2_high1<=0)|(sigma2_high2<=0)
        comb_bounds_low[violated],comb_bounds_high[violated] = 0,1
        return comb_bounds_low,comb_bounds_high
    
    def dependency(self, comb_flags):
        return np.repeat(comb_flags[:,None],3,axis=1)



from .abstract_cub_mlmc import AbstractCubMLMC
from ..util.data import Data

from ..discrete_distribution import IIDStdUniform
from ..discrete_distribution.abstract_discrete_distribution import AbstractIIDDiscreteDistribution
from ..true_measure import Gaussian
from ..integrand import FinancialOption
from ..util import MaxSamplesWarning, ParameterError, MaxLevelsWarning, ParameterWarning
import numpy as np
from scipy.stats import norm
from time import time
import warnings


class CubMLMC(AbstractCubMLMC):
    """
    Multilevel IID Monte Carlo stopping criterion.
    
    Examples:
        >>> fo = FinancialOption(IIDStdUniform(seed=7))
        >>> sc = CubMLMC(fo,abs_tol=1.5e-2)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.785
            n_total         3577556
            levels          2^(2)
            n_level         [2438191  490331  207606   62905]
            mean_level      [1.715 0.053 0.013 0.003]
            var_level       [21.829  1.761  0.452  0.11 ]
            cost_per_sample [ 2.  4.  8. 16.]
            alpha           2.008
            beta            1.997
            gamma           1.000
            time_integrate  ...
        CubMLMC (AbstractStoppingCriterion)
            rmse_tol        0.006
            n_init          2^(8)
            levels_min      2^(1)
            levels_max      10
            theta           2^(-1)
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        1
            drift           0
            mean            0
            covariance      1
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               1
            replications    1
            entropy         7

    **References:**
        
    1.  M.B. Giles. 'Multi-level Monte Carlo path simulation'.   
        Operations Research, 56(3):607-617, 2008.  
        [http://people.maths.ox.ac.uk/~gilesm/files/OPRE_2008.pdf](http://people.maths.ox.ac.uk/~gilesm/files/OPRE_2008.pdf).

    2. [http://people.maths.ox.ac.uk/~gilesm/mlmc/#MATLAB](http://people.maths.ox.ac.uk/~gilesm/mlmc/#MATLAB).
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = .05, 
                 rmse_tol = None, 
                 n_init = 256, 
                 n_limit = 1e10, 
                 alpha = .01, 
                 levels_min = 2, 
                 levels_max = 10, 
                 alpha0 = -1., 
                 beta0 = -1., 
                 gamma0 = -1.,
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rmse_tol (np.ndarray): Root mean squared error tolerance. 
                If supplied, then absolute tolerance and alpha are ignored in favor of the rmse tolerance. 
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            levels_min (int): Minimum level of refinement $\geq 2$.
            levels_max (int): Maximum level of refinement $\geq$ `levels_min`.
            alpha0 (float): Weak error is $\mathcal{O}(2^{-\alpha_0\ell})$ in the level $\ell$. If `alpha0`$\leq 0$ then it will be estimated. 
            beta0 (float): Variance is $\mathcal{O}(2^{-\beta_0\ell})$ in the level $\ell$. If `beta0`$\leq 0$ then it will be estimated. 
            gamma0 (float): Sample cost is $\mathcal{O}(2^{\gamma_0\ell})$ in the level $\ell$. If `gamma0`$\leq 0$ then it will be estimated. 
        """
        self.parameters = ['rmse_tol','n_init','levels_min','levels_max','theta']
        if levels_min < 2:
            raise ParameterError('needs levels_min >= 2')
        if levels_max < levels_min:
            raise ParameterError('needs levels_max >= levels_min')
        if n_init <= 0:
            raise ParameterError('needs n_init > 0')
        # initialization
        if rmse_tol:
            self.rmse_tol = float(rmse_tol)
        else: # use absolute tolerance
            self.rmse_tol =  float(abs_tol) / norm.ppf(1-alpha/2)
        self.alpha = alpha 
        assert 0<self.alpha<1
        self.n_init = n_init
        self.n_limit = n_limit
        self.levels_min = levels_min
        self.levels_max = levels_max
        self.theta = 0.5
        self.alpha0 = alpha0
        self.beta0 = beta0
        self.gamma0 = gamma0
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        super(CubMLMC,self).__init__(allowed_distribs=[AbstractIIDDiscreteDistribution], allow_vectorized_integrals=False)

    def integrate(self):
        t_start = time()
        data = Data(parameters=[
            'solution',
            'n_total',
            'levels',
            'n_level',
            'mean_level',
            'var_level', 
            'cost_per_sample',
            'alpha',
            'beta',
            'gamma'])
        data.levels = int(self.levels_min)
        data.n_level = np.zeros(data.levels+1,dtype=int)
        data.sum_level = np.zeros((2,data.levels+1))
        data.cost_level = np.zeros(data.levels+1)
        data.diff_n_level = self.n_init*np.ones(data.levels+1,dtype=int)
        data.alpha = np.maximum(0,self.alpha0)
        data.beta = np.maximum(0,self.beta0)
        data.gamma = np.maximum(0,self.gamma0)
        data.level_integrands = []
        while data.diff_n_level.sum() > 0:
            self._update_data(data)
            data.n_total += data.diff_n_level.sum()
            # set optimal number of additional samples
            n_samples = self._get_next_samples(data)
            data.diff_n_level = np.maximum(0, n_samples-data.n_level)
            # check if over sample budget
            if (data.n_total + data.diff_n_level.sum()) > self.n_limit:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples, which would exceed n_limit = %d.
                Stopping integration process.
                Note that error tolerances may no longer be satisfied""" \
                % (int(data.n_total), int(data.diff_n_level.sum()), int(self.n_limit))
                warnings.warn(warning_s, MaxSamplesWarning)
                break
            # if (almost) converged, estimate remaining error and decide 
            # whether a new level is required
            if (data.diff_n_level > 0.01*data.n_level).sum() == 0:
                range_ = np.arange(min(2.,data.levels-1)+1)
                idx = (data.levels-range_).astype(int)
                rem = (data.mean_level[idx] / 
                        2.**(range_*data.alpha)).max() / (2.**data.alpha - 1)
                # rem = ml(l+1) / (2^alpha - 1)
                if rem > np.sqrt(self.theta)*self.rmse_tol:
                    if data.levels == self.levels_max:
                        warnings.warn(
                            'Failed to achieve weak convergence. levels == levels_max.',
                            MaxLevelsWarning)
                    else:
                        self._add_level(data)
                        n_samples = self._get_next_samples(data)
                        data.diff_n_level = np.maximum(0,n_samples-data.n_level)
        # finally, evaluate multilevel estimator
        data.solution = (data.sum_level[0,:]/data.n_level).sum()
        data.levels += 1
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data



import numpy as np
import torch 
import gpytorch

class ExactGPyTorchRegressionModel(gpytorch.models.ExactGP):
    allowed_likelihood_types = (
        gpytorch.likelihoods.GaussianLikelihood,
        gpytorch.likelihoods.GaussianLikelihoodWithMissingObs,
        gpytorch.likelihoods.FixedNoiseGaussianLikelihood)
    def __init__(self, x_t, y_t, prior_mean, prior_cov, likelihood, use_gpu=False):
        if isinstance(x_t,np.ndarray): x_t = torch.from_numpy(x_t)
        if isinstance(y_t,np.ndarray): y_t = torch.from_numpy(y_t)
        assert x_t.ndim==2 and y_t.ndim==1 and len(x_t)==len(y_t)
        super(ExactGPyTorchRegressionModel, self).__init__(x_t,y_t,likelihood)
        assert isinstance(self.likelihood,ExactGPyTorchRegressionModel.allowed_likelihood_types)
        self.mean_module,self.covar_module = prior_mean,prior_cov
        self.d = x_t.shape[1]
        self.use_gpu = use_gpu
        if self.use_gpu:
            assert torch.cuda.is_available()
            self = self.cuda()
            self.likelihood = self.likelihood.cuda()
    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x,covar_x)
    def fit(self, optimizer, mll, training_iter, verbose=0):
        self.train()
        self.likelihood.train()
        if verbose: print('\tgpytorch model fitting')
        for i in range(training_iter):
            optimizer.zero_grad()
            output = self.__call__(self.train_inputs[0])
            loss = -mll(output,self.train_targets)
            loss.backward()
            if verbose and (i+1)%verbose==0:
                print('\t\titer %-3d of %d'%(i+1,training_iter))
                for name,val in self.named_parameters(): print('\t\t\t%s %.2e'%(name.ljust(50,'.'),val))
            optimizer.step()
    def predict(self, x, noise_const=0, chunk_size=2**15):
        if isinstance(x,np.ndarray): x = torch.from_numpy(x)
        assert x.ndim==2 and x.shape[1]==self.d
        self.eval()
        self.likelihood.eval()
        n = len(x)
        mean_post,std_post = np.zeros(n,dtype=float),np.zeros(n,dtype=float)
        for i in range(0,n,chunk_size):
            lchunk,uchunk = i,min(i+chunk_size,n)
            noise_chunk = noise_const*torch.ones(uchunk-lchunk)
            mean_post[lchunk:uchunk],std_post[lchunk:uchunk] = self._predict_batch(x[lchunk:uchunk],noise_chunk)
        return mean_post,std_post
    def _predict_batch(self, x, noise):
        if self.use_gpu: x,noise = x.cuda(),noise.cuda()
        with torch.no_grad(): observed_pred = self.likelihood(self.__call__(x),noise=noise)
        mean_post = observed_pred.mean
        std_post = observed_pred.stddev
        if self.use_gpu: mean_post,std_post = mean_post.cpu(),std_post.cpu()
        if self.use_gpu: del x,noise,observed_pred; torch.cuda.empty_cache()
        return mean_post.numpy(),std_post.numpy()
    def add_data(self, x_t_new, y_t_new):
        if isinstance(x_t_new,np.ndarray): x_t_new = torch.from_numpy(x_t_new)
        if isinstance(y_t_new,np.ndarray): y_t_new = torch.from_numpy(y_t_new)
        assert x_t_new.ndim==2 and x_t_new.shape[1]==self.d and y_t_new.ndim==1 and len(x_t_new)==len(y_t_new)
        if self.use_gpu: x_t_new,y_t_new = x_t_new.cuda(),y_t_new.cuda()
        fantasy_model = self.get_fantasy_model(x_t_new,y_t_new)
        if self.use_gpu: del x_t_new,y_t_new; torch.cuda.empty_cache()
        return fantasy_model


from .abstract_true_measure import AbstractTrueMeasure
from ..util import DimensionError
from ..discrete_distribution import DigitalNetB2
import numpy as np
from scipy.stats import norm
from typing import Union


class Uniform(AbstractTrueMeasure):
    r"""
    Uniform distribution, see [https://en.wikipedia.org/wiki/Continuous_uniform_distribution](https://en.wikipedia.org/wiki/Continuous_uniform_distribution). 

    Examples:
        >>> true_measure = Uniform(DigitalNetB2(2,seed=7),lower_bound=[0,.5],upper_bound=[2,3])
        >>> true_measure(4)
        array([[1.44324713, 2.7873875 ],
               [0.32691107, 1.5741214 ],
               [1.97352511, 0.58590959],
               [0.8591331 , 1.89690854]])
        >>> true_measure
        Uniform (AbstractTrueMeasure)
            lower_bound     [0.  0.5]
            upper_bound     [2 3]

        With independent replications 

        >>> x = Uniform(DigitalNetB2(3,seed=7,replications=2),lower_bound=[.25,.5,.75],upper_bound=[1.75,1.5,1.25])(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[0.61979915, 0.6821862 , 1.12366296],
                [1.27229355, 1.16169442, 0.9644598 ],
                [0.97209782, 1.29818233, 0.79100643],
                [1.62311988, 0.79520621, 1.13747905]],
        <BLANKLINE>
               [[0.92315337, 1.35899604, 1.0027484 ],
                [1.05453886, 0.54353443, 0.91782473],
                [0.59821215, 0.79281506, 0.78420518],
                [1.37943573, 1.10241448, 1.13481488]]])
    """
    
    def __init__(self, sampler, lower_bound=0, upper_bound=1):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            lower_bound (Union[float,np.ndarray]): Lower bound.
            upper_bound (Union[float,np.ndarray]): Upper bound.
        """
        self.parameters = ['lower_bound', 'upper_bound']
        self.domain = np.array([[0,1]])
        self._parse_sampler(sampler)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        if np.isscalar(self.lower_bound):
            lower_bound = np.tile(self.lower_bound,self.d)
        if np.isscalar(self.upper_bound):
            upper_bound = np.tile(self.upper_bound,self.d)
        self.a = np.array(lower_bound)
        self.b = np.array(upper_bound)
        if len(self.a)!=self.d or len(self.b)!=self.d:
            raise DimensionError('upper bound and lower bound must be of length dimension')
        self.delta = self.b - self.a
        self.inv_delta_prod = 1/self.delta.prod()
        self.range = np.hstack((self.a.reshape((self.d,1)),self.b.reshape((self.d,1))))
        super(Uniform,self).__init__()
        assert self.a.shape==(self.d,) and self.b.shape==(self.d,)

    def _transform(self, x):
        return x*self.delta+self.a
    
    def _weight(self, x):
        return np.tile(self.inv_delta_prod,x.shape[:-1])
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = Uniform(sampler,lower_bound=self.a,upper_bound=self.b)
        else:
            l = self.a[0]
            u = self.b[0]
            if not (all(self.a==l) and all(self.b==u)):
                raise DimensionError('''
                    In order to spawn a uniform measure with different dimensions
                    the lower bounds must all be the same and 
                    the upper bounds must all be the same''')
            spawn = Uniform(sampler,lower_bound=l,upper_bound=u)
        return spawn
    



from .abstract_stopping_criterion import AbstractStoppingCriterion
import numpy as np
from scipy.stats import norm


class AbstractCubMLQMC(AbstractStoppingCriterion):
    
    def set_tolerance(self, abs_tol=None, rel_tol=None, rmse_tol=None):
        assert rel_tol is None, "rel_tol not supported by this stopping criterion."
        if rmse_tol != None:
            self.rmse_tol = float(rmse_tol)
        elif abs_tol != None:
            self.rmse_tol = (float(abs_tol) / norm.ppf(1-self.alpha/2.))
    
    def update_data(self, data):
        # update sample sums
        for l in range(data.levels):
            if not data.eval_level[l]:
                # nothing to do on this level
                continue
            if l==len(data.level_integrands):
                # haven't spawned this level's integrand yet
                data.level_integrands += self.integrand.spawn(levels=l)
            # reset dimension
            n_max = self.n_init if data.n_level[l]==0 else 2*data.n_level[l]
            integrand_l = data.level_integrands[l]
            n_min = data.n_level[l]
            samples = integrand_l.discrete_distrib(n_min=n_min,n_max=n_max)
            n = n_max-n_min
            pc,pf = integrand_l.f(samples).squeeze()
            dp = pf-pc
            prev_sum = data.mean_level_reps[l]*data.n_level[l]
            data.mean_level_reps[l] = (dp.sum(-1)+prev_sum)/float(n_max)
            data.cost_level[l] = data.cost_level[l] + self.replications*n*integrand_l.cost
            data.n_level[l] = n_max
            data.mean_level[l] = data.mean_level_reps[l].mean()
            data.var_level[l] = data.mean_level_reps[l].var()
            cost_per_sample = data.cost_level[l]/data.n_level[l]/self.replications
            data.var_cost_ratio_level[l] = data.var_level[l]/cost_per_sample
        self._update_bias_estimate(data)
        data.n_total = self.replications*data.n_level.sum()
        data.solution = data.mean_level.sum()
        data.eval_level[:] = False # Reset active levels

    def _update_bias_estimate(self, data):
        A = np.ones((2,2))
        A[:,0] = range(data.levels-2, data.levels)
        y = np.ones(2)
        y[0] = np.log2(abs(data.mean_level_reps[data.levels-2].mean()))
        y[1] = np.log2(abs(data.mean_level_reps[data.levels-1].mean()))
        x = np.linalg.lstsq(A,y,rcond=None)[0]
        alpha = max(.5,-x[0])
        data.bias_estimate = 2**(x[1]+data.levels*x[0]) / (2**alpha - 1)
    
    def _add_level(self, data):
        # Add another level to relevant attributes.
        data.levels += 1
        if data.levels > len(data.n_level):
            data.n_level = np.hstack((data.n_level,0))
            data.eval_level = np.hstack((data.eval_level,True))
            data.mean_level_reps = np.vstack((data.mean_level_reps,np.zeros(int(self.replications))))
            data.mean_level = np.hstack((data.mean_level,0))
            data.var_level = np.hstack((data.var_level,np.inf))
            data.cost_level = np.hstack((data.cost_level,0))
            data.var_cost_ratio_level = np.hstack((data.var_cost_ratio_level,np.inf))

    def _add_level_MLMC(self, data):
        # Add another level to relevant attributes.
        data.levels += 1
        if not len(data.n_level) > data.levels:
            data.mean_level = np.hstack((data.mean_level, data.mean_level[-1] / 2**data.alpha))
            data.var_level = np.hstack((data.var_level, data.var_level[-1] / 2**data.beta))
            data.cost_per_sample = np.hstack((data.cost_per_sample, data.cost_per_sample[-1] * 2**data.gamma))
            data.n_level = np.hstack((data.n_level, 0.))
            data.sum_level = np.hstack((data.sum_level,np.zeros((2,1))))
            data.cost_level = np.hstack((data.cost_level, 0.))
        else:
            data.mean_level = np.absolute(data.sum_level[0,:data.levels+1]/data.n_level[:data.levels+1])
            data.var_level = np.maximum(0,data.sum_level[1,:data.levels+1]/data.n_level[:data.levels+1] - data.mean_level**2)
            data.cost_per_sample = data.cost_level[:data.levels+1]/data.n_level[:data.levels+1]



import numpy as np
from .abstract_integrand import AbstractIntegrand
from ..true_measure import Uniform
from ..discrete_distribution import DigitalNetB2

class Sin1d(AbstractIntegrand):
    r"""
    Sine function in $d=1$ dimension.  

    $$g(t) = \sin(t), \qquad t \sim \mathcal{U}[0,2\pi k]$$

    Examples:
        >>> integrand = Sin1d(DigitalNetB2(1,seed=7),k=1)
        >>> y = integrand(2**10)
        >>> print("%.4e"%y.mean())
        -1.3582e-10
        >>> integrand.true_measure
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     6.283
        
        With independent replications

        >>> integrand = Sin1d(DigitalNetB2(1,seed=7,replications=2**4),k=1)
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4e"%muhats.mean())
        7.0800e-04
    """
    def __init__(self, sampler, k=1):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            k (float): The true measure will be uniform between $0$ and $2 \pi k$.
        """
        self.sampler = sampler
        self.k = k
        assert self.sampler.d==1
        self.true_measure = Uniform(self.sampler,lower_bound=0,upper_bound=2*self.k*np.pi)
        super(Sin1d,self).__init__(dimension_indv=(),dimension_comb=(),parallel=False)
    def g(self, t):
        return np.sin(t[...,0])
    def _spawn(self, level, sampler):
        return Sin1d(sampler=sampler,k=self.k)



def _load_c_lib():
    from ctypes import CDLL, RTLD_GLOBAL
    from os.path import dirname, abspath
    from glob import glob
    import os
    _c_lib = None
    for file in glob(dirname(abspath(__file__)) + os.sep + "_c_lib*"):
        try:
            _c_lib = CDLL(file, mode=RTLD_GLOBAL)
            break
        except OSError:
            pass
    assert isinstance(_c_lib,CDLL), "problem importing QMCPy's C library"
    return _c_lib


from .abstract_integrand import AbstractIntegrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Uniform
import numpy as np


class BoxIntegral(AbstractIntegrand):
    r"""
    Box integral from [1], see also 
    
    $$B_s(\boldsymbol{t}) = \left(\sum_{j=1}^d t_j^2 \right)^{s/2}, \qquad \boldsymbol{T} \sim \mathcal{U}[0,1]^d.$$

    Examples:
        Scalar `s` 
        
        >>> integrand = BoxIntegral(DigitalNetB2(2,seed=7),s=7)
        >>> y = integrand(2**10)
        >>> y.shape
        (1024,)
        >>> print("%.4f"%y.mean(0))
        0.7519

        With independent replications

        >>> integrand = BoxIntegral(DigitalNetB2(2,seed=7,replications=2**4),s=7)
        >>> y = integrand(2**10)
        >>> y.shape
        (16, 1024)
        >>> muhats = y.mean(1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean(0))
        0.7518

        Array `s`

        >>> integrand = BoxIntegral(DigitalNetB2(5,seed=7),s=np.arange(6).reshape((2,3)))
        >>> y = integrand(2**10)
        >>> y.shape
        (2, 3, 1024)
        >>> y.mean(-1)
        array([[1.        , 1.26234461, 1.66666661],
               [2.28201516, 3.22195096, 4.67188113]])

        With independent replications

        >>> integrand = BoxIntegral(DigitalNetB2(2,seed=7,replications=2**4),s=np.arange(6).reshape((2,3)))
        >>> y = integrand(2**10)
        >>> y.shape
        (2, 3, 16, 1024)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (2, 3, 16)
        >>> muhats.mean(-1)
        array([[1.        , 0.76519118, 0.66666666],
               [0.62718785, 0.62224086, 0.64273341]])

    **References:**

    1.  D.H. Bailey, J.M. Borwein, R.E. Crandall, Box integrals.  
        Journal of Computational and Applied Mathematics, Volume 206, Issue 1, 2007, Pages 196-208, ISSN 0377-0427.  
        [https://doi.org/10.1016/j.cam.2006.06.010](https://doi.org/10.1016/j.cam.2006.06.010).  
        [https://www.sciencedirect.com/science/article/pii/S0377042706004250](https://www.sciencedirect.com/science/article/pii/S0377042706004250).  
        [https://www.davidhbailey.com/dhbpapers/boxintegrals.pdf](https://www.davidhbailey.com/dhbpapers/boxintegrals.pdf)
    """

    def __init__(self, sampler, s=1):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            s (Union[float,np.ndarray]): `s` parameter or parameters. The output shape of `g` is the shape of `s`. 
        """
        self.parameters = ['s']
        self.s = np.array(s)
        assert self.s.size>0
        self.sampler = sampler
        self.true_measure = Uniform(self.sampler)
        self.s_over_2 = self.s/2
        super(BoxIntegral,self).__init__(dimension_indv=self.s.shape,dimension_comb=self.s.shape,parallel=False)

    def g(self, t, **kwargs):
        sum_squares = (t**2).sum(-1)
        y = sum_squares**self.s_over_2[(...,)+(None,)*sum_squares.ndim]
        return y
    
    def _spawn(self, level, sampler):
        return BoxIntegral(sampler=sampler,s=self.s)



from ..util import MethodImplementationError, _univ_repr, ParameterError
from ..true_measure.abstract_true_measure import AbstractTrueMeasure
from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
import numpy as np
import os
from multiprocessing import get_context
from multiprocessing.pool import ThreadPool
from itertools import repeat


class AbstractIntegrand(object):

    def __init__(self, dimension_indv, dimension_comb, parallel, threadpool=False):
        r"""
        Args:
            dimension_indv (tuple): Individual solution shape.
            dimension_comb (tuple): Combined solution shape. 
            parallel (int): Parallelization flag. 
                
                - When `parallel = 0` or `parallel = 1` then function evaluation is done in serial fashion.
                - `parallel > 1` specifies the number of processes used by `multiprocessing.Pool` or `multiprocessing.pool.ThreadPool`.
            
                Setting `parallel=True` is equivalent to `parallel = os.cpu_count()`.
            threadpool (bool): When `parallel > 1`: 
                
                - Setting `threadpool = True` will use `multiprocessing.pool.ThreadPool`.
                - Setting `threadpool = False` will use `setting multiprocessing.Pool`.
        """
        prefix = 'A concrete implementation of AbstractIntegrand must have '
        self.d = self.true_measure.d
        self.d_indv = (dimension_indv,) if isinstance(dimension_indv,int) else tuple(dimension_indv)
        self.d_comb = (dimension_comb,) if isinstance(dimension_comb,int) else tuple(dimension_comb)
        cpus = os.cpu_count()
        self.parallel = cpus if parallel is True else int(parallel)
        self.parallel = 0 if self.parallel==1 else self.parallel
        self.threadpool = threadpool
        if not (hasattr(self, 'sampler') and isinstance(self.sampler,(AbstractTrueMeasure,AbstractDiscreteDistribution))):
            raise ParameterError(prefix + 'self.sampler, a AbstractTrueMeasure or AbstractDiscreteDistribution instance')
        if not (hasattr(self, 'true_measure') and isinstance(self.true_measure,AbstractTrueMeasure)):
            raise ParameterError(prefix + 'self.true_measure, a AbstractTrueMeasure instance')
        if not hasattr(self,'parameters'):
            self.parameters = []
        if not hasattr(self,'multilevel'):
            self.multilevel = False
        assert isinstance(self.multilevel,bool)
        if not hasattr(self,'max_level'):
            self.max_level = np.inf
        if not hasattr(self,'discrete_distrib'):
            self.discrete_distrib = self.true_measure.discrete_distrib
        if self.true_measure.transform!=self.true_measure and \
           not (self.true_measure.range==self.true_measure.transform.range).all():
            raise ParameterError("The range of the composed transform is not compatible with this true measure")
        self.EPS = np.finfo(float).eps

    def __call__(self, n=None, n_min=None, n_max=None, warn=True):
        r"""
        - If just `n` is supplied, generate samples from the sequence at indices 0,...,`n`-1.
        - If `n_min` and `n_max` are supplied, generate samples from the sequence at indices `n_min`,...,`n_max`-1.
        - If `n` and `n_min` are supplied, then generate samples from the sequence at indices `n`,...,`n_min`-1.

        Args:
            n (Union[None,int]): Number of points to generate.
            n_min (Union[None,int]): Starting index of sequence.
            n_max (Union[None,int]): Final index of sequence.
            warn (bool): If `False`, disable warnings when generating samples.

        Returns:
            t (np.ndarray): Samples from the sequence. 
                
                - If `replications` is `None` then this will be of size (`n_max`-`n_min`) $\times$ `dimension` 
                - If `replications` is a positive int, then `t` will be of size `replications` $\times$ (`n_max`-`n_min`) $\times$ `dimension` 
            weights (np.ndarray): Only returned when `return_weights=True`. The Jacobian weights for the transformation
        """
        return self.gen_samples(n=n,n_min=n_min,n_max=n_max,warn=warn)
    
    def gen_samples(self, n=None, n_min=None, n_max=None, return_weights=False, warn=True):
        x = self.discrete_distrib(n=n,n_min=n_min,n_max=n_max,warn=warn)
        y = self.f(x)
        return y

    def g(self, t, *args, **kwargs):
        r"""
        *Abstract method* implementing the integrand as a function of the true measure.

        Args:
            t (np.ndarray): Inputs with shape `(*batch_shape, d)`.
            args (tuple): positional arguments to `g`.
            kwargs (dict): keyword arguments to `g`. 
                
                Some algorithms will additionally try to pass in a `compute_flags` keyword argument. 
                This `np.ndarray` are flags indicating which outputs require evaluation.  
                For example, if the vector function has 3 outputs and `compute_flags = [False, True, False]`, 
                then the function is only required to evaluate the second output and may leave the remaining outputs as `np.nan` values, 
                i.e., the outputs corresponding to `compute_flags` which are `False` will not be used in the computation.

        Returns:
            y (np.ndarray): function evaluations with shape `(*batch_shape, *dimension_indv)` where `dimension_indv` is the shape of the function outputs. 
        """
        raise MethodImplementationError(self, 'g')

    def f(self, x, *args, **kwargs):
        r"""
        Function to evaluate the transformed integrand as a function of the discrete distribution.  
        Automatically applies the transformation determined by the true measure. 

        Args:
            x (np.ndarray): Inputs with shape `(*batch_shape, d)`.
            args (tuple): positional arguments to `g`.
            kwargs (dict): keyword arguments to `g`. 
                
                Some algorithms will additionally try to pass in a `compute_flags` keyword argument. 
                This `np.ndarray` are flags indicating which outputs require evaluation.  
                For example, if the vector function has 3 outputs and `compute_flags = [False, True, False]`, 
                then the function is only required to evaluate the second output and may leave the remaining outputs as `np.nan` values, 
                i.e., the outputs corresponding to `compute_flags` which are `False` will not be used in the computation.
                
                The keyword argument `periodization_transform`, a string, specifies a periodization transform. 
                Options are: 

                - `False`: No periodizing transform, $\psi(x) = x$. 
                - `'BAKER'`: Baker tansform $\psi(x) = 1-2\lvert x-1/2 \rvert$.
                - `'C0'`: $C^0$ transform $\psi(x) = 3x^2-2x^3$.
                - `'C1'`: $C^1$ transform $\psi(x) = x^3(10-15x+6x^2)$.
                - `'C1SIN'`: Sidi $C^1$ transform $\psi(x) = x-\sin(2 \pi x)/(2 \pi)$. 
                - `'C2SIN'`: Sidi $C^2$ transform $\psi(x) = (8-9 \cos(\pi x)+\cos(3 \pi x))/16$.
                - `'C3SIN'`: Sidi $C^3$ transform $\psi(x) = (12\pi x-8\sin(2 \pi x) + \sin(4 \pi x))/(12 \pi)$.

        Returns:
            y (np.ndarray): function evaluations with shape `(*batch_shape, *dimension_indv)` where `dimension_indv` is the shape of the function outputs. 
        """
        if "periodization_transform" in kwargs:
            periodization_transform = kwargs["periodization_transform"]
            del kwargs["periodization_transform"]
        else:
            periodization_transform = "None"
        periodization_transform = str(periodization_transform).upper()
        batch_shape = tuple(x.shape[:-1])
        d_indv_ndim = len(self.d_indv)
        if periodization_transform=="NONE": periodization_transform = "FALSE"
        if self.discrete_distrib.mimics != 'StdUniform' and periodization_transform!='NONE':
            raise ParameterError('''
                Applying a periodization transform currently requires a discrete distribution 
                that mimics a standard uniform measure.''')
        if periodization_transform == 'FALSE':
            xp = x
            wp = np.ones(batch_shape,dtype=float)
        elif periodization_transform == 'BAKER':
            xp = 1-2*abs(x-1/2)
            wp = np.ones(batch_shape,dtype=float)
        elif periodization_transform == 'C0':
            xp = 3*x**2-2*x**3
            wp = np.prod(6*x*(1-x),-1)
        elif periodization_transform == 'C1':
            xp = x**3*(10-15*x+6*x**2)
            wp = np.prod(30*x**2*(1-x)**2,-1)
        elif periodization_transform == 'C1SIN':
            xp = x - np.sin(2*np.pi*x)/(2*np.pi)
            wp = np.prod(2*np.sin(np.pi*x)**2,-1)
        elif periodization_transform == 'C2SIN':
            xp = (8-9*np.cos(np.pi*x)+np.cos(3*np.pi*x))/16
            wp = np.prod((9*np.sin(np.pi*x)*np.pi-np.sin(3*np.pi*x)*3*np.pi)/16,-1)
        elif periodization_transform=='C3SIN':
            xp = (12*np.pi*x-8*np.sin(2*np.pi*x)+np.sin(4*np.pi*x))/(12*np.pi)
            wp = np.prod((12*np.pi-8*np.cos(2*np.pi*x)*2*np.pi+np.sin(4*np.pi*x)*4*np.pi)/(12*np.pi),-1)
        else:
            raise ParameterError("The %s periodization transform is not implemented"%periodization_transform)
        if periodization_transform in ['C1','C1SIN','C2SIN','C3SIN']:
            xp[xp<=0] = self.EPS
            xp[xp>=1] = 1-self.EPS
        assert wp.shape==batch_shape
        assert xp.shape==x.shape
        # function evaluation with chain rule
        i = (None,)*d_indv_ndim+(...,)
        if self.true_measure==self.true_measure.transform:
            # jacobian*weight/pdf will cancel so f(x) = g(\Psi(x))
            xtf = self.true_measure._jacobian_transform_r(xp,return_weights=False) # get transformed samples, equivalent to self.true_measure._transform_r(x)
            assert xtf.shape==xp.shape
            y = self._g(xtf,*args,**kwargs)
        else: # using importance sampling --> need to compute pdf, jacobian(s), and weight explicitly
            pdf = self.discrete_distrib.pdf(xp) # pdf of samples
            assert pdf.shape==batch_shape
            xtf,jacobians = self.true_measure.transform._jacobian_transform_r(xp,return_weights=True) # compute recursive transform+jacobian
            assert xtf.shape==xp.shape 
            assert jacobians.shape==batch_shape
            weight = self.true_measure._weight(xtf) # weight based on the true measure
            assert weight.shape==batch_shape
            gvals = self._g(xtf,*args,**kwargs)
            assert gvals.shape==(self.d_indv+batch_shape)
            y = gvals*weight[i]/pdf[i]*jacobians[i]
        assert y.shape==(self.d_indv+batch_shape)
        # account for periodization weight
        y = y*wp[i]
        assert y.shape==(self.d_indv+batch_shape)
        return y

    def _g(self, t, *args, **kwargs):
        if self.parallel:
            pool = get_context(method='fork').Pool(processes=self.parallel) if self.threadpool==False else ThreadPool(processes=self.parallel) 
            t_flat = t.reshape((-1,t.shape[-1]))
            y_flat = pool.starmap(self._g2,zip(t_flat,repeat((args,kwargs))))
            pool.close()
            y = np.stack(y_flat,axis=-1).reshape((self.d_indv+t.shape[:-1]))
        else:
            y = self._g2(t,comb_args=(args,kwargs))
        expected_y_shape = (self.d_indv+t.shape[:-1])
        assert y.shape==expected_y_shape, "expected y.shape to be %s but got %s"%(str(expected_y_shape),str(y.shape))
        return y

    def _g2(self, t, comb_args=((),{})):
        args = comb_args[0]
        kwargs = comb_args[1]
        try:
            y = self.g(t,*args,**kwargs)
        except TypeError as e:
            if "got an unexpected keyword argument 'compute_flags'" in str(e):
                del kwargs['compute_flags']
                y = self.g(t,*args,**kwargs)
            else: raise e
        return y

    def bound_fun(self, bound_low, bound_high):
        """
        Compute the bounds on the combined function based on bounds for the
        individual functions.  

        Defaults to the identity where we essentially
        do not combine integrands, but instead integrate each function
        individually.

        Args:
            bound_low (np.ndarray): Lower bounds on individual estimates with shape `integrand.d_indv`.
            bound_high (np.ndarray): Upper bounds on individual estimates with shape `integrand.d_indv`.

        Returns:
            comb_bound_low (np.ndarray): Lower bounds on combined estimates with shape `integrand.d_comb`.
            comb_bound_high (np.ndarray): Upper bounds on combined estimates with shape `integrand.d_comb`.
        """
        if self.d_indv!=self.d_comb:
            raise ParameterError('''
                Set bound_fun explicitly. 
                The default bound_fun is the identity map. 
                Since the individual solution dimensions d_indv = %s does not equal the combined solution dimensions d_comb = %d, 
                QMCPy cannot infer a reasonable bound function.'''%(str(self.d_indv),str(self.d_comb)))
        return bound_low,bound_high

    def dependency(self, comb_flags):
        """
        Takes a vector of indicators of weather of not the error bound is satisfied for combined integrands and returns flags for individual integrands.  

        For example, if we are taking the ratio of 2 individual integrands, then getting `comb_flags=True` means the ratio
        has not been approximated to within the tolerance, so the dependency function should return `indv_flags=[True,True]`
        indicating that both the numerator and denominator integrands need to be better approximated.

        Args:
            comb_flags (np.ndarray): Flags of shape `integrand.d_comb` indicating whether the combined outputs are insufficiently approximated.

        Returns:
            indv_flags (np.ndarray): Flags of shape `integrand.d_indv` indicating whether the individual integrands require additional sampling. 
        """
        return comb_flags if self.d_indv==self.d_comb else np.tile((comb_flags==False).any(),self.d_indv)

    def spawn(self, levels):
        r"""
        Spawn new instances of the current integrand at different levels with new seeds.
        Used by multi-level QMC algorithms which require integrands at multiple levels.

        Note:
            Use `replications` instead of using `spawn` when possible, e.g., when spawning copies which all have the same level.

        Args:
            levels (np.ndarray): Levels at which to spawn new integrands.
        
        Returns:
            spawned_integrand (list): Integrands with new true measures and discrete distributions.
        """
        levels = np.array([levels]) if np.isscalar(levels) else np.array(levels)
        if (levels>self.max_level).any():
            raise ParameterError("requested spawn level exceeds max level")
        n_levels = len(levels)
        new_dims = np.array([self.dimension_at_level(level) for level in levels])
        tm_spawns = self.sampler.spawn(s=n_levels,dimensions=new_dims) 
        spawned_integrand = [None]*n_levels
        for l,level in enumerate(levels):
            spawned_integrand[l] = self._spawn(level,tm_spawns[l])
        return spawned_integrand
    
    def dimension_at_level(self, level):
        """
        *Abstract method* which returns the dimension of the generator required for a given level.

        Note:
            Only used for multilevel problems.

        Args:
            level (int): Level at which to return the dimension.

        Returns:
            d (int): Dimension at the given input level. 
        """
        return self.d

    def _spawn(self, level, tm_spawn):
        raise MethodImplementationError(self, '_spawn')

    def __repr__(self):
        return _univ_repr(self, "AbstractIntegrand", self.parameters)



from ..abstract_discrete_distribution import AbstractLDDiscreteDistribution
from ...util import ParameterError, ParameterWarning
import qmctoolscl
import numpy as np
from numpy.lib.npyio import DataSource
from os.path import dirname, abspath, isfile
import warnings
from copy import deepcopy
from typing import Union


class Lattice(AbstractLDDiscreteDistribution):
    r"""
    Low discrepancy lattice sequence.

    Note:
        - Lattice sample sizes should be powers of $2$ e.g. $1$, $2$, $4$, $8$, $16$, $\dots$.
        - The first point of an unrandomized lattice is the origin.
    
    Examples:
        >>> discrete_distrib = Lattice(2,seed=7)
        >>> discrete_distrib(4)
        array([[0.04386058, 0.58727432],
               [0.54386058, 0.08727432],
               [0.29386058, 0.33727432],
               [0.79386058, 0.83727432]])
        >>> discrete_distrib(1) # first point in the sequence
        array([[0.04386058, 0.58727432]])
        >>> discrete_distrib
        Lattice (AbstractLDDiscreteDistribution)
            d               2^(1)
            replications    1
            randomize       SHIFT
            gen_vec_source  kuo.lattice-33002-1024-1048576.9125.txt
            order           RADICAL INVERSE
            n_limit         2^(20)
            entropy         7

        Replications of independent randomizations

        >>> x = Lattice(3,seed=7,replications=2)(4)
        >>> x.shape
        (2, 4, 3)
        >>> x
        array([[[0.04386058, 0.58727432, 0.3691824 ],
                [0.54386058, 0.08727432, 0.8691824 ],
                [0.29386058, 0.33727432, 0.1191824 ],
                [0.79386058, 0.83727432, 0.6191824 ]],
        <BLANKLINE>
               [[0.65212985, 0.69669968, 0.10605352],
                [0.15212985, 0.19669968, 0.60605352],
                [0.90212985, 0.44669968, 0.85605352],
                [0.40212985, 0.94669968, 0.35605352]]])


        Different orderings (avoid warnings that the first point is the origin).

        >>> Lattice(dimension=2,randomize=False,order='RADICAL INVERSE')(4,warn=False) 
        array([[0.  , 0.  ],
               [0.5 , 0.5 ],
               [0.25, 0.75],
               [0.75, 0.25]])
        >>> Lattice(dimension=2,randomize=False,order='GRAY')(4,warn=False)
        array([[0.  , 0.  ],
               [0.5 , 0.5 ],
               [0.75, 0.25],
               [0.25, 0.75]])
        >>> Lattice(dimension=2,randomize=False,order='LINEAR')(4,warn=False)
        array([[0.  , 0.  ],
               [0.25, 0.75],
               [0.5 , 0.5 ],
               [0.75, 0.25]])

        Generating vector from [https://github.com/QMCSoftware/LDData/tree/main/lattice](https://github.com/QMCSoftware/LDData/tree/main/lattice)

        >>> Lattice(dimension=3,randomize=False,generating_vector="mps.exod2_base2_m20_CKN.txt")(8,warn=False)
        array([[0.   , 0.   , 0.   ],
               [0.5  , 0.5  , 0.5  ],
               [0.25 , 0.75 , 0.75 ],
               [0.75 , 0.25 , 0.25 ],
               [0.125, 0.375, 0.375],
               [0.625, 0.875, 0.875],
               [0.375, 0.125, 0.125],
               [0.875, 0.625, 0.625]])
        
        Random generating vector supporting $2^{25}$ points 

        >>> discrete_distrib = Lattice(3,generating_vector=25,seed=55,randomize=False)
        >>> discrete_distrib.gen_vec
        array([[       1, 11961679, 12107519]], dtype=uint64)
        >>> discrete_distrib(4,warn=False)
        array([[0.  , 0.  , 0.  ],
               [0.5 , 0.5 , 0.5 ],
               [0.25, 0.75, 0.75],
               [0.75, 0.25, 0.25]])
        
        Two random generating vectors both supporting $2^{25}$ points along with independent random shifts

        >>> discrete_distrib = Lattice(3,seed=7,generating_vector=25,replications=2)
        >>> discrete_distrib.gen_vec
        array([[       1, 32809149,  1471719],
               [       1,   275319, 19705657]], dtype=uint64)
        >>> discrete_distrib(4)
        array([[[0.3691824 , 0.65212985, 0.69669968],
                [0.8691824 , 0.15212985, 0.19669968],
                [0.6191824 , 0.90212985, 0.44669968],
                [0.1191824 , 0.40212985, 0.94669968]],
        <BLANKLINE>
               [[0.10605352, 0.63025643, 0.13630282],
                [0.60605352, 0.13025643, 0.63630282],
                [0.35605352, 0.38025643, 0.38630282],
                [0.85605352, 0.88025643, 0.88630282]]])
    
    **References**
    
    1.  Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama, Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou.  
        GAIL: Guaranteed Automatic Integration Library (Version 2.3), MATLAB Software, 2019.  
        [http://gailgithub.github.io/GAIL_Dev/](http://gailgithub.github.io/GAIL_Dev/).  

    2.  F.Y. Kuo, D. Nuyens.  
        Application of quasi-Monte Carlo methods to elliptic PDEs with random diffusion coefficients \- a survey of analysis and implementation.  
        Foundations of Computational Mathematics, 16(6):1631-1696, 2016.  
        [https://link.springer.com/article/10.1007/s10208-016-9329-5](https://link.springer.com/article/10.1007/s10208-016-9329-5).  

    3.  D. Nuyens.  
        The Magic Point Shop of QMC point generators and generating vectors.  
        MATLAB and Python software, 2018.  
        [https://people.cs.kuleuven.be/~dirk.nuyens/](https://people.cs.kuleuven.be/~dirk.nuyens/).

    4.  R. Cools, F.Y. Kuo, D. Nuyens.  
        Constructing embedded lattice rules for multivariate integration.  
        SIAM J. Sci. Comput., 28(6), 2162-2188.

    5.  P. L'Ecuyer, D. Munger.  
        LatticeBuilder: A General Software Tool for Constructing Rank-1 Lattice Rules.  
        ACM Transactions on Mathematical Software. 42. (2015).  
        [10.1145/2754929](https://dl.acm.org/doi/10.1145/2754929). 
    """

    def __init__(self,
                 dimension = 1,
                 replications = None,
                 seed = None,
                 randomize = 'SHIFT',
                 generating_vector = "kuo.lattice-33002-1024-1048576.9125.txt",
                 order = 'RADICAL INVERSE',
                 m_max = None):
        r"""
        Args:
            dimension (Union[int,np.ndarray]): Dimension of the generator.

                - If an `int` is passed in, use generating vector components at indices 0,...,`dimension`-1.
                - If an `np.ndarray` is passed in, use generating vector components at these indices.
            
            replications (int): Number of independent randomizations.
            seed (Union[None,int,np.random.SeedSeq): Seed the random number generator for reproducibility.
            randomize (str): Options are 
                
                - `'SHIFT'`: Random shift.
                - `'FALSE'`: No randomization. In this case the first point will be the origin. 
            
            generating_vector (Union[str,np.ndarray,int]: Specify the generating vector.
                
                - A `str` should be the name (or path) of a file from the LDData repo at [https://github.com/QMCSoftware/LDData/tree/main/lattice](https://github.com/QMCSoftware/LDData/tree/main/lattice).
                - A `np.ndarray` of integers with shape $(d,)$ or $(r,d)$ where $d$ is the number of dimensions and $r$ is the number of replications.  
                    Must supply `m_max` where $2^{m_\mathrm{max}}$ is the max number of supported samples. 
                - An `int`, call it $M$, 
                gives the random generating vector $(1,v_1,\dots,v_{d-1})^T$ 
                where $d$ is the dimension and $v_i$ are randomly selected from $\{3,5,\dots,2^M-1\}$ uniformly and independently.  
                We require require $1 < M < 27$. 

            order (str): `'LINEAR'`, `'RADICAL INVERSE'`, or `'GRAY'` ordering. See the doctest example above.
            m_max (int): $2^{m_\mathrm{max}}$ is the maximum number of supported samples.
        """
        self.parameters = ['randomize','gen_vec_source','order','n_limit']
        self.input_generating_vector = deepcopy(generating_vector)
        self.input_m_max = deepcopy(m_max)
        if isinstance(generating_vector,str) and generating_vector=="kuo.lattice-33002-1024-1048576.9125.txt":
            self.gen_vec_source = generating_vector
            gen_vec = np.load(dirname(abspath(__file__))+'/generating_vectors/kuo.lattice-33002-1024-1048576.9125.npy')[None,:]
            d_limit = 9125
            n_limit = 1048576
        elif isinstance(generating_vector,str):
            self.gen_vec_source = generating_vector
            assert generating_vector[-4:]==".txt"
            local_root = dirname(abspath(__file__))+'/generating_vectors/'
            repos = DataSource()
            if repos.exists(local_root+generating_vector):
                datafile = repos.open(local_root+generating_vector)
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/lattice/"+generating_vector):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/lattice/"+generating_vector)
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/"+generating_vector):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/"+generating_vector)
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/lattice/"+generating_vector[7:]):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/lattice/"+generating_vector[7:])
            elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/"+generating_vector):
                datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/"+generating_vector)
            elif repos.exists(generating_vector):
                datafile = repos.open(generating_vector)
            else:
                raise ParameterError("LDData path %s not found"%generating_vector)
            contents = [int(line.rstrip('\n').strip().split("#",1)[0]) for line in datafile.readlines() if line[0]!="#"]
            datafile.close()
            d_limit = int(contents[0])
            n_limit = int(contents[1])
            gen_vec = np.array(contents[2:],dtype=np.uint64)[None,:]
        elif isinstance(generating_vector,np.ndarray):
            self.gen_vec_source = "custom"
            gen_vec = generating_vector
            if m_max is None:
                raise ParameterError("m_max must be supplied when generating_vector is a np.ndarray")
            n_limit = int(2**m_max)
            d_limit = int(gen_vec.shape[-1])
        elif isinstance(generating_vector,int):
            assert 1<generating_vector<27, "int generating vector out of range"
            n_limit = 2**generating_vector
            assert isinstance(dimension,int), "random generating vector requires int dimension"
            d_limit = dimension
        else:
            raise ParameterError("invalid generating_vector, must be a string, numpy.ndarray, or int")
        super(Lattice,self).__init__(dimension,replications,seed,d_limit,n_limit)
        if isinstance(generating_vector,int):
            self.gen_vec_source = "random"
            m_max = int(np.log2(self.n_limit))
            gen_vec = np.hstack([np.ones((self.replications,1),dtype=np.uint64),2*self.rng.integers(1,2**(m_max-1),size=(self.replications,dimension-1),dtype=np.uint64)+1]).copy()
        assert isinstance(gen_vec,np.ndarray)
        gen_vec = np.atleast_2d(gen_vec) 
        assert gen_vec.ndim==2 and gen_vec.shape[1]>=self.d and (gen_vec.shape[0]==1 or gen_vec.shape[0]==self.replications), "invalid gen_vec.shape = %s"%str(gen_vec.shape)
        self.gen_vec = gen_vec[:,self.dvec].copy()
        self.order = str(order).upper().strip().replace("_"," ")
        if self.order=="GRAY CODE": self.order = "GRAY"
        if self.order=="NATURAL": self.order = "RADICAL INVERSE"
        assert self.order in ['LINEAR','RADICAL INVERSE','GRAY']
        self.randomize = str(randomize).upper()
        if self.randomize=="TRUE": self.randomize = "SHIFT"
        if self.randomize=="NONE": self.randomize = "FALSE"
        if self.randomize=="NO": self.randomize = "FALSE"
        assert self.randomize in ["SHIFT","FALSE"]
        if self.randomize=="SHIFT":
            self.shift = self.rng.uniform(size=(self.replications,self.d))
        if self.randomize=="FALSE": assert self.gen_vec.shape[0]==self.replications, "randomize='FALSE' but replications = %d does not equal the number of sets of generating vectors %d"%(self.replications,self.gen_vec.shape[0])
        

    def _gen_samples(self, n_min, n_max, return_binary, warn):
        if return_binary:
            raise ParameterError("Lattice does not support return_binary=True")
        if n_min==0 and self.randomize=="FALSE" and warn:
            warnings.warn("Without randomization, the first lattice point is the origin",ParameterWarning)
        r_x = np.uint64(self.gen_vec.shape[0]) 
        n = np.uint64(n_max-n_min)
        d = np.uint64(self.d) 
        n_start = np.uint64(n_min)
        x = np.empty((r_x,n,d),dtype=np.float64)
        if self.order=="LINEAR":
            assert r_x==1, "lattice linear currently requires there be only 1 generating matrix"
            x = self._gail_linear(n_min,n_max)[None,:,:]
        elif self.order=="RADICAL INVERSE":
            assert (n_min==0 or (n_min&(n_min-1))==0) and (n_max==0 or (n_max&(n_max-1))==0), "lattice in natural order requires n_min and n_max be 0 or powers of 2"
            _ = qmctoolscl.lat_gen_natural(r_x,n,d,n_start,self.gen_vec,x,backend="c")
        elif self.order=="GRAY":
            _ = qmctoolscl.lat_gen_gray(r_x,n,d,n_start,self.gen_vec,x,backend="c") 
        else: 
            assert False, "invalid lattice order"
        if self.randomize=="FALSE":
            xr = x
        elif self.randomize=="SHIFT":
            r = np.uint64(self.replications)
            xr = np.empty((r,n,d),dtype=np.float64)
            qmctoolscl.lat_shift_mod_1(r,n,d,r_x,x,self.shift,xr,backend="c")
        return xr

    def _gen_block_linear(self, m_next, first=True):
        n = int(2**m_next)
        if first:
            y = np.arange(0, 1, 1 / n).reshape((n, 1))
        else:
            y = np.arange(1 / n, 1, 2 / n).reshape((n, 1))
        x = np.outer(y, self.gen_vec) % 1
        return x

    def calculate_y(self, m_low, m_high, y):
        for m in range(m_low, m_high):
            n = 2 ** m
            y_next = np.arange(1 / n, 1, 2 / n).reshape((int(n / 2), 1))
            temp = np.zeros((n, 1))
            temp[0::2] = y
            temp[1::2] = y_next
            y = temp
        return y

    def _gail_linear(self, n_min, n_max):
        """ Gail lattice generator in linear order. """
        m_low = int(np.floor(np.log2(n_min))) + 1 if n_min > 0 else 0
        m_high = int(np.ceil(np.log2(n_max)))
        if n_min == 0:
            return self._gen_block_linear(m_high, first=True)
        else:
            n = 2 ** (m_low)
            y = np.arange(1 / n, 1, 2 / n).reshape((int(n / 2), 1))
            y = self.calculate_y(m_low, m_high, y)
            x = np.outer(y, self.gen_vec) % 1
            return x

    def _spawn(self, child_seed, dimension):
        return Lattice(
                dimension = dimension,
                replications = None if self.no_replications else self.replications,
                seed = child_seed,
                randomize = self.randomize,
                generating_vector = self.input_generating_vector,
                order = self.order,
                m_max = self.input_m_max,
                )



from .abstract_true_measure import AbstractTrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution import DigitalNetB2
import numpy as np
from numpy.linalg import cholesky, slogdet
from scipy.stats import norm, multivariate_normal
from scipy.linalg import eigh
from typing import Union


class Gaussian(AbstractTrueMeasure):
    """
    Gaussian (Normal) distribution as described in [https://en.wikipedia.org/wiki/Multivariate_normal_distribution](https://en.wikipedia.org/wiki/Multivariate_normal_distribution).
    
    Note:
        - `Normal` is an alias for `Gaussian`
    
    Examples:
        >>> true_measure = Gaussian(DigitalNetB2(2,seed=7),mean=[1,2],covariance=[[9,4],[4,5]])
        >>> true_measure(4)
        array([[ 3.83994612,  1.19097885],
               [-1.9727727 ,  0.49405353],
               [ 5.87242307,  8.41341485],
               [ 0.61222205,  1.48402653]])
        >>> true_measure
        Gaussian (AbstractTrueMeasure)
            mean            [1 2]
            covariance      [[9 4]
                             [4 5]]
            decomp_type     PCA
        
        With independent replications 

        >>> x = Gaussian(DigitalNetB2(3,seed=7,replications=2),mean=0,covariance=3)(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[-1.18721904, -1.57108272,  1.15371635],
                [ 0.81749123,  0.72242445, -0.31025434],
                [-0.0807895 ,  1.44651585, -2.41042379],
                [ 2.38133494, -0.93225637,  1.30817519]],
        <BLANKLINE>
               [[-0.22304017,  1.86337427,  0.02386568],
                [ 0.15807672, -2.96365385, -0.73502346],
                [-1.26753687, -0.94427848, -2.57683314],
                [ 1.1844196 ,  0.44964332,  1.27760936]]])
    """

    def __init__(self, sampler, mean=0., covariance=1., decomp_type='PCA'):
        """
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            mean (Union[float,np.ndarray]): Mean vector. 
            covariance (Union[float,np.ndarray]): Covariance matrix. A float or vector will be expanded into a diagonal matrix.  
            decomp_type (str): Method for decomposition for covariance matrix. Options include
             
                - `'PCA'` for principal component analysis, or 
                - `'Cholesky'` for cholesky decomposition.
        """
        self.parameters = ['mean', 'covariance', 'decomp_type']
        # default to transform from standard uniform
        self.domain = np.array([[0,1]])
        self._parse_sampler(sampler)
        self._parse_gaussian_params(mean,covariance,decomp_type)
        self.range = np.array([[-np.inf,np.inf]])
        super(Gaussian,self).__init__()
        assert self.mu.shape==(self.d,) and self.a.shape==(self.d,self.d)
    
    def _parse_gaussian_params(self, mean, covariance, decomp_type):
        self.decomp_type = decomp_type.upper()
        self.mean = mean
        self.covariance = covariance
        if np.isscalar(mean):
            mean = np.tile(mean,self.d)
        if np.isscalar(covariance):
            covariance = covariance*np.eye(self.d)
        self.mu = np.array(mean)
        self.sigma = np.array(covariance)
        if self.sigma.shape==(self.d,):
            self.sigma = np.diag(self.sigma)
        self.sigma = (self.sigma+self.sigma.T)/2
        if not (len(self.mu)==self.d and self.sigma.shape==(self.d,self.d)):
            raise DimensionError('''
                    mean must have length d and
                    covariance must be of shape d x d''')
        if self.decomp_type == 'PCA':
            evals,evecs = eigh(self.sigma) # get eigenvectors and eigenvalues for
            evecs = evecs*(1-2*(evecs[0]<0)) # force first entries of eigenvectors to be positive
            order = np.argsort(-evals)
            self.a = np.dot(evecs[:,order],np.diag(np.sqrt(evals[order])))
        elif self.decomp_type == 'CHOLESKY':
            self.a = cholesky(self.sigma)
        else:
            raise ParameterError("decomp_type should be 'PCA' or 'Cholesky'") 
        self.mvn_scipy = multivariate_normal(mean=self.mu,cov=self.sigma, allow_singular=True)

    def _transform(self, x):
        return self.mu+np.einsum("...ij,kj->...ik",norm.ppf(x),self.a)

    def _weight(self, t):
        return self.mvn_scipy.pdf(t)
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = Gaussian(sampler,mean=self.mu,covariance=self.covariance,decomp_type=self.decomp_type)
        else:
            m = self.mu[0]
            c = self.sigma[0,0]
            expected_cov = c*np.eye(int(self.d))
            if not ( (self.mu==m).all() and (self.sigma==expected_cov).all() ):
                raise DimensionError('''
                        In order to spawn a Gaussian measure
                        mean (mu) must be all the same and 
                        covariance must be a scaler times I''')
            spawn = Gaussian(sampler,mean=m,covariance=c,decomp_type=self.decomp_type)
        return spawn
